from django.utils.timezone import now
from salesforce import models
from salesforce.models import SalesforceModel


class User(SalesforceModel):  # type: ignore[type-arg]
    name = models.CharField(max_length=80)


class Account(SalesforceModel):  # type: ignore[type-arg]
    name = models.CharField(max_length=80)
    owner = models.ForeignKey(User, on_delete=models.DO_NOTHING, default=models.DEFAULTED_ON_CREATE,)


class Contact(SalesforceModel):  # type: ignore[type-arg]
    last_name = models.CharField(max_length=40)
    donor_class = models.CharField(custom=True, db_column='Donor_class__c', max_length=255, verbose_name='Donor class',
                                   default=models.DefaultedOnCreate('None'), blank=True, null=False)


class Unreal(SalesforceModel):  # type: ignore[type-arg]
    bool_x = models.BooleanField(default=models.DEFAULTED_ON_CREATE)
    int_x = models.IntegerField(default=models.DEFAULTED_ON_CREATE)
    float_x = models.FloatField(default=models.DEFAULTED_ON_CREATE)
    decimal_x = models.DecimalField(default=models.DEFAULTED_ON_CREATE, decimal_places=2, max_digits=9)
    str_x = models.CharField(default=models.DEFAULTED_ON_CREATE, max_length=255)
    date_x = models.DateField(default=models.DEFAULTED_ON_CREATE)
    datetime_x = models.DateTimeField(default=models.DEFAULTED_ON_CREATE)
    time_x = models.TimeField(default=models.DEFAULTED_ON_CREATE)
    str2_x = models.CharField(default=models.DefaultedOnCreate('no'), max_length=255)
    bool2_x = models.BooleanField(default=models.DefaultedOnCreate(True))
    callable_timestamp = models.DateTimeField(default=models.DefaultedOnCreate(now))
    normal = models.CharField(max_length=255)



"""Test the completness ad validity of tooling inspectdb by read and write all tables.

It tries to find model and read one record from all retrieveable tables, except
objects with extraordinal filter conditions. Then it tries to write this
record back if the table is updateable, except some tables. This can fail
on permissions.

Usage:
$ python manage.py inspectdb --database=salesforce --tooling-api --settings=tests.tooling.settings \
        >tests/tooling/models.py
$ python manage.py check --settings=tests.tooling.settings
$ python tests/tooling/slow_test.py
"""

from inspect import isclass
from sys import stdout, stderr
import os
import sys

import django
sys.path.insert(0, '.')
os.environ['DJANGO_SETTINGS_MODULE'] = 'tests.tooling.settings'

django.setup()

# these 3 lines must be imported after: path, environ, django.setup()
from django.db import connections  # noqa
from tests.tooling import models as mdl  # noqa
from salesforce.backend.base import SalesforceError  # noqa


# see https://developer.salesforce.com/docs/atlas.en-us.api_tooling.meta/api_tooling/reference_objects_soql_limits.htm
# "The following objects in Tooling API don’t support SOQL operations
#  COUNT(), GROUP BY, LIMIT, LIMIT OFFSET, OR, NOT, and INCLUDES."
soql_operation_limitation = [
    'CompactLayoutInfo', 'CompactLayoutItemInfo', 'DataType', 'EntityDefinition',
    'EntityLimit', 'EntityParticle', 'FieldDefinition', 'Publisher', 'RelationshipDomain',
    'RelationshipInfo', 'SearchLayout', 'ServiceFieldDataType', 'StandardAction',
    'TimeSheetTemplate', 'UserEntityAccess', 'UserFieldAccess',
]

sf = connections['salesforce']


def run():
    start_name = sys.argv[1] if sys.argv[1:] else ''
    sf.introspection.is_tooling_api = True
    n_tables = n_read = n_no_data = n_read_errors = n_write = n_write_errors = 0
    model_map = {cls._meta.db_table: cls
                 for cls in mdl.__dict__.values()
                 if isclass(cls) and issubclass(cls, django.db.models.Model)
                 }
    problematic_read = {
        # document in a private folder
        'EmailTemplate',
        # only queries of the form Id='<some_value>'
        'SubscriberPackage', 'SubscriberPackageVersion',

        # "When retrieving results with Metadata or FullName fields, the query
        #  qualifications must specify no more than one row for retrieval."
        'EntityDefinition',
        # invalid Metadata JSON
        'OpportunitySettings',
        # any reason
        'AIReplyRecommendationsSettings', 'AccountInsightsSettings', 'ActivitiesSettings',
        'ApexExecutionOverlayResult', 'AutomatedContactsSettings', 'BusinessHoursEntry',
        'CaseClassificationSettings', 'CompactLayoutInfo', 'CompactLayoutItemInfo',
        'DataDotComSettings', 'EinsteinAssistantSettings', 'EmailToCaseRoutingAddress',
        'EntityLimit', 'FieldServiceSettings', 'ForecastingObjectListSelectedSettings',
        'ForecastingObjectListSettings', 'FormulaVariable', 'HighVelocitySalesSettings',
        'IndustriesManufacturingSettings', 'KnowledgeAnswerSettings', 'KnowledgeCaseField',
        'KnowledgeCaseSettings', 'KnowledgeLanguageSettings', 'KnowledgeSettings',
        'KnowledgeSuggestedArticlesSettings', 'KnowledgeWorkOrderField',
        'KnowledgeWorkOrderLineItemField', 'LiveAgentSettings', 'MenuItem',
        'ObjectSearchSetting', 'OpportunityInsightsSettings',
        'OpportunityListFieldsLabelMapping', 'OpportunityScoreSettings',
        'OrderManagementSettings', 'StandardAction', 'StandardValueSet',
        # in API 50.0 Winter'20
        'InventorySettings', 'SecuritySettings',
        # in API 51.0 Spring '21
        'EmployeeUserSettings',
        # in API 52.0 Summer '21
        'EinsteinAgentSettings'
    }
    problematic_write = {
        # any SalesforceError
        'AccountIntelligenceSettings', 'AccountSettings', 'ActionsSettings',
        'AnalyticsSettings', 'ApexClass', 'ApexPage', 'ApexSettings', 'ApexTrigger',
        'AppExperienceSettings', 'BlockchainSettings', 'BusinessHoursSettings',
        'BusinessProcess', 'CampaignSettings', 'CaseSettings', 'ChatterEmailsMDSettings',
        'ChatterSettings', 'CommunitiesSettings', 'ConnectedAppSettings', 'ContentSettings',
        'CurrencySettings', 'DashboardMobileSettings', 'DeploymentSettings', 'EACSettings',
        'EmailAdministrationSettings', 'EmailIntegrationSettings', 'EmailTemplateSettings',
        'EnhancedNotesSettings', 'EventSettings', 'FilesConnectSettings', 'FlowSettings',
        'ForecastingSettings', 'ForecastingTypeSettings', 'FormulaSettings',
        'GoogleAppsSettings', 'IndustriesSettings', 'InvocableActionSettings',
        'IsvHammerSettings', 'LanguageSettings', 'LeadConfigSettings', 'LeadConvertSettings',
        'LightningExperienceSettings', 'MacroSettings', 'ManagedContentType',
        'MobileSettings', 'MyDomainSettings', 'NotificationsSettings', 'PardotSettings',
        'PartyDataModelSettings', 'PathAssistantSettings', 'PicklistSettings',
        'PlatformEncryptionSettings', 'PortalsSettings', 'PrivacySettings', 'ProductSettings',
        'QuickTextSettings', 'RecentlyViewed', 'RecordPageSettings', 'SchemaSettings',
        'SearchSettings', 'SecuritySettings', 'SharingSettings', 'SocialProfileSettings',
        'SurveySettings', 'SystemNotificationSettings', 'Territory2Settings',
        'TrialOrgSettings', 'User', 'UserEngagementSettings', 'UserInterfaceSettings',
        'UserManagementSettings', 'WebToXSettings', 'WorkDotComSettings',

        # in API 50.0 Winter'20
        'ExternalServicesSettings', 'RecommendationBuilderSettings',

        # ExpirationDate must be in the future.
        'TraceFlag',
        # silently not updated
        'CustomApplication', 'CustomField', 'CustomTab', 'FlexiPage', 'GlobalValueSet',
        'Layout', 'QuickActionDefinition', 'RemoteProxy', 'ValidationRule',
        'WorkflowFieldUpdate', 'WorkflowRule',
        'QuickActionListItem',
    }

    #  FIELD_INTEGRITY_EXCEPTION
    # Only the Metadata and FullName fields may be specified on AddressSettings, or else Metadata must be excluded.

    requires_filter = [
        # some of them are important, but can not be tested automatically here
        'ColorDefinition', 'EntityParticle', 'FieldDefinition', 'IconDefinition',
        'OwnerChangeOptionInfo', 'RelationshipDomain', 'RelationshipInfo', 'SearchLayout',
        'SiteDetail', 'UserEntityAccess', 'UserFieldAccess',
    ]
    new_problematic = []
    sf.connect()
    # cur = sf.connection.cursor()
    for tab in sf.introspection.table_list_cache['sobjects']:
        db_table = tab['name']
        if tab['queryable'] and (db_table in model_map and db_table not in requires_filter and
                                 db_table not in problematic_read):
            if db_table < start_name:
                continue
            test_class = model_map[db_table]
            stdout.write('%s' % db_table)
            obj = None
            try:
                n_read += 1
                # cur.execute("SELECT Id FROM {} LIMIT 1".format(db_table), tooling_api=True)
                obj = test_class.objects.all()[0]
            except SalesforceError as exc:
                new_problematic.append(db_table)
                stderr.write("\n************** %s %s\n" % (db_table, exc))
                n_read_errors += 1
            except IndexError:
                n_no_data += 1
            if obj:
                stdout.write("* ")
            if obj and tab['updateable'] and db_table not in problematic_write:
                stdout.write('(write) ')
                try:
                    n_write += 1
                    obj.save(force_update=True)
                except SalesforceError as exc:
                    new_problematic.append(db_table)
                    stderr.write("\n************** write SalesforceError %s %s\n" % (db_table, exc))
                    n_write_errors += 1
                except (TypeError, NotImplementedError) as exc:
                    new_problematic.append(db_table)
                    stderr.write("\n************** write %s %s %s\n" % (exc.__class__.__name__, db_table, exc))
                    n_write_errors += 1
                else:
                    # object 'Topic' doesn't have the attribute 'last_modified_date'
                    # in recently created SFDC databases (proably version 34.0+)
                    if hasattr(obj, 'last_modified_date'):
                        if not (test_class.objects.get(pk=obj.pk).last_modified_date > obj.last_modified_date):
                            new_problematic.append(db_table)
                            print("not updated - -")
                            n_write_errors += 1
            stdout.write('\n')
    if new_problematic:
        print("\nnew_problematic:\n%r\n" % new_problematic)
    n_tables = len(sf.introspection.table_list_cache['sobjects'])
    print('Tooling api results:')
    print('  {n_tables} tables, {n_read} reads tried, {n_no_data} no data, '
          '{n_read_errors} read errors, {n_write} writes tried, {n_write_errors} write errors'
          .format(n_tables=n_tables, n_read=n_read, n_no_data=n_no_data,
                  n_read_errors=n_read_errors, n_write=n_write, n_write_errors=n_write_errors))
    print('********* ERRORs found' if n_read_errors + n_write_errors else 'OK')
    return n_read_errors + n_write_errors == 0


if __name__ == '__main__':
    ok = run()
    sys.exit(0 if ok else 1)



from salesforce import models


class Contact(models.SalesforceModel):
    last_name = models.CharField(max_length=80)
    # a field that is not used in example.Contact
    title = models.CharField(max_length=40, blank=True, null=True)

    class Meta(models.Model.Meta):
        db_table = 'Contact'
        verbose_name = 'Contact'
        verbose_name_plural = 'Contacts'
        app_label = 'test_salesforce'



"""Backward compatible behaviour with primary key 'Id'."""
from django.test import TestCase
from tests.test_compatibility.models import Lead, B, AtoB
from salesforce.backend.test_helpers import current_user, uid_version as uid


class CompatibilityTest(TestCase):
    databases = '__all__'

    def test_capitalized_id(self):
        test_lead = Lead(Company='sf_test lead' + uid, LastName='name')
        test_lead.save()
        try:
            refreshed_lead = Lead.objects.get(Id=test_lead.Id)
            self.assertEqual(refreshed_lead.Id, test_lead.Id)
            self.assertEqual(refreshed_lead.Owner.Username, current_user)
            leads = Lead.objects.filter(Company='sf_test lead' + uid, LastName='name')
            self.assertGreaterEqual(len(leads), 1)
            repr(test_lead.__dict__)
        finally:
            test_lead.delete()


class DjangoCompatibility(TestCase):
    databases = '__all__'

    def test_autofield_compatible(self):
        """Test that the light weigh AutoField is compatible in all Django ver."""
        primary_key = [x for x in Lead._meta.fields if x.primary_key][0]
        self.assertEqual(primary_key.auto_created, True)
        self.assertEqual(primary_key.get_internal_type(), 'AutoField')
        self.assertIn(primary_key.name, ('id', 'Id'))


class ManyToManyTest(TestCase):

    def test_subquery_filter_on_child(self):
        """Filter with a Subquery() on a child object.

        Especially useful with ManyToMany field relationships.
        """
        assoc = AtoB.objects.filter(a__email='a@example.com')
        qs = B.objects.filter(pk__in=assoc.values('b'))
        soql = (
            "SELECT B__c.Id FROM B__c WHERE B__c.Id IN ("
            "SELECT AtoB__c.B__c FROM AtoB__c WHERE AtoB__c.A__r.Email__c = 'a@example.com'"
            ")"
        )
        self.assertEqual(str(qs.query), soql)



"""Demonstrate that a Model can inherite from more abstract models."""

from django.conf import settings
import salesforce
from salesforce import models
from salesforce.models import SalesforceModel

# All demo models simplified for readability, except tested features


class User(SalesforceModel):
    username = models.CharField(max_length=80)
    email = models.CharField(max_length=100)


class DefaultMixin(SalesforceModel):
    """Common fields used in the most of SFDC models."""
    last_modified_date = models.DateTimeField(sf_read_only=models.READ_ONLY, auto_now=True)
    owner = models.ForeignKey(User, on_delete=models.DO_NOTHING,
                              default=models.DEFAULTED_ON_CREATE)  # db_column='OwnerId'

    class Meta:
        abstract = True


class CommonAccount(DefaultMixin, SalesforceModel):
    """Common fields of Salesforce Account model."""
    description = models.TextField()
    phone = models.CharField(max_length=255)

    class Meta:
        abstract = True


class CoreAccount(SalesforceModel):
    """Fields specific to standard Account only."""
    name = models.CharField(max_length=255)

    class Meta:
        abstract = True


class PersonAccount(SalesforceModel):
    """Fields specific to Account after activating "Person Account"."""
    LastName = models.CharField(max_length=80)
    FirstName = models.CharField(max_length=40)
    Name = models.CharField(max_length=255, sf_read_only=models.READ_ONLY)
    IsPersonAccount = models.BooleanField(default=False, sf_read_only=models.READ_ONLY)
    PersonEmail = models.CharField(max_length=100)

    class Meta:
        abstract = True


if not getattr(settings, 'SF_EXAMPLE_PERSON_ACCOUNT_ACTIVATED', False):
    class Account(CommonAccount, CoreAccount):
        pass
else:
    class Account(CommonAccount, PersonAccount):  # type: ignore[no-redef] # noqa
        pass


class DummyMixin:
    def some_overridden_method(self):
        pass


class DummyMixin2:
    pass


class Contact(DummyMixin, DefaultMixin, SalesforceModel, DummyMixin2):
    name = models.CharField(max_length=255, sf_read_only=models.READ_ONLY)
    last_name = models.CharField(max_length=80)
    first_name = models.CharField(max_length=40, blank=True)
    account = salesforce.fields.ForeignKey(Account, on_delete=salesforce.models.DO_NOTHING)


class ProxyContact(Contact):
    class Meta:
        proxy = True


class Proxy2Contact(ProxyContact):
    class Meta:
        proxy = True



from django.db import connections
from django.test import TestCase
import requests

from salesforce.backend.base import SalesforceError
from salesforce.testrunner.example.models import User


class DynamicAuthTest(TestCase):
    databases = '__all__'

    def assertConnectionProblem(self, queryset):
        with self.assertRaises(Exception) as cm:
            len(queryset)
        exc = cm.exception
        self.assertTrue(isinstance(exc, (requests.ConnectionError, SalesforceError)))

    def test_dynamic_auth(self):
        """
        Verify dynamic auth: Get access token from a backup 'salesforce_copy' of the normal connection.
        The dynamic connection 'salesforce' is without credentials.
        Verify that exception is raised before dynamic_start(...) or after dynamic_end().
        """
        users = User.objects.all()[:2]
        # get access_token from the 'salesforce_copy' connection.
        with connections['salesforce_copy'].cursor() as cursor:
            access_token = cursor.oauth['access_token']
            instance_url = cursor.oauth['instance_url']
        # verify that uncofigured dynamic-auth 'salesforce' fails initially
        self.assertConnectionProblem(users)
        # dynamic auth succeeds after getting the token
        connections['salesforce'].sf_session.auth.dynamic_start(access_token, instance_url=instance_url)
        self.assertGreater(len(users), 0)
        # verify that it fails again after 'dynamic_end()'
        connections['salesforce'].sf_session.auth.dynamic_end()
        users = User.objects.all()[:2]
        self.assertConnectionProblem(users)



"""A writable test, not specifically important what is tested"""
from django.db import transaction
from django.db.utils import IntegrityError
from django.test import TestCase

from .models import Lead, User


class Test(TestCase):
    databases = '__all__'

    def test_lead_owner_without_default(self):
        test_user = User.objects.create(username='user', last_name='a', email='a@a.au')
        test_lead = Lead(company='sf_test lead', last_name='name')
        try:
            with self.assertRaises(IntegrityError) as cm, transaction.atomic():
                # can't be saved without owner
                test_lead.save()
            self.assertIn('NOT NULL constraint failed', cm.exception.args[0])

            # can be saver with owner
            test_lead.owner = test_user
            test_lead.save()
        finally:
            if test_lead.id is not None:
                test_lead.delete()
        test_user.delete()



from django.test import TestCase
from tests.test_mixin.models import Account, Contact, Proxy2Contact
from salesforce.backend.test_helpers import current_user, uid_version as uid


def refresh(obj):
    """
    Get the same object refreshed from the same db.
    """
    db = obj._state.db
    return type(obj).objects.using(db).get(pk=obj.pk)


class MixinTest(TestCase):
    databases = '__all__'

    def test_mixin(self):
        """Test that mixins from abstract classes work and also proxy models."""
        # create the object with one field from the second ancestor and one from the first
        test_account = Account(name='sf_test account' + uid, description='experimental')
        test_account.save()
        test_account = refresh(test_account)
        test_contact = Contact(first_name='sf_test', last_name='my', account=test_account)
        test_contact.save()
        try:
            test_contact = refresh(test_contact)
            # verify foreign keys to and from the complicated model
            self.assertEqual(test_contact.account.owner.username, current_user)
            contacts = Contact.objects.filter(account__name='sf_test account' + uid)  # description='experimental')
            self.assertGreaterEqual(len(contacts), 1)
            repr(test_contact.__dict__)
            repr(test_account.__dict__)
            # Verify that a proxy model correctly recognizes the db_table from
            # a concrete model by two levels deeper.
            self.assertEqual(Proxy2Contact.objects.get(pk=test_contact.pk).pk, test_contact.pk)
        finally:
            test_contact.delete()
            test_account.delete()



"""
A writable test, not specifically important what is tested

both databases are non-salesforce
"""
from django.conf import settings
from django.test import TestCase

from salesforce.backend.test_helpers import default_is_sf
from .models import Contact, User


class CompatibilityTest(TestCase):
    databases = '__all__'

    def test_capitalized_id(self):
        # prepare a Contact
        test_contact = Contact(last_name='Smith')
        if not default_is_sf:
            # If SALESFORCE_DB_ALIAS is not a Salesforce database then a user must be created
            # and set as the owner, because DefaultedOnCreate works only on SFDC servers
            test_user = User.objects.create(username='test', last_name='test', email='test@example.com')
            test_contact.owner = test_user
        test_contact.save()

        # test that the primary key 'Id' works
        sf_pk = getattr(settings, 'SF_PK', 'id')
        self.assertTrue(hasattr(test_contact, sf_pk))
        try:
            refreshed_contact = Contact.objects.get(pk=test_contact.pk)
            self.assertEqual(refreshed_contact.pk, test_contact.pk)
        finally:
            test_contact.delete()



import unittest
from salesforce.backend.test_helpers import current_user
from tests.tooling.models import (
    # CustomObject, CustomField,
    EntityDefinition, FieldDefinition,
    UserEntityAccess, UserFieldAccess,
    # PermissionDependency,
    # PermissionSet, PermissionSetGroup, PermissionSetGroupComponent,
    User,
)


class PermissionTest(unittest.TestCase):
    def test_permissions_by_tooling(self):
        # "When retrieving results with Metadata or FullName fields, the query
        #  qualificatio​ns must specify no more than one row for retrieval."
        entity_definitions = EntityDefinition.objects.defer('full_name', 'metadata')
        entity_definition = EntityDefinition.objects.get(qualified_api_name='Contact')
        assert entity_definitions

        field_definitions = FieldDefinition.objects.filter(qualified_api_name='Contact').defer('full_name', 'metadata')
        field_definitions = (FieldDefinition.objects.filter(entity_definition_id=entity_definition.durable_id)
                             .defer('full_name', 'metadata'))
        field_definition = FieldDefinition.objects.get(entity_definition_id='Contact', qualified_api_name='Email')
        assert field_definitions

        current_user_id = User.objects.get(username=current_user).pk

        user_entity_access = UserEntityAccess.objects.get(durable_id='%s.%s' %
                                                          (entity_definition.durable_id, current_user_id))
        user_field_access = UserFieldAccess.objects.get(durable_id='%s.%s' %
                                                        (field_definition.durable_id, current_user_id))

        x = user_entity_access
        assert [x.is_editable, x.is_readable] == [True, True]

        x = user_field_access
        assert [x.is_accessible, x.is_creatable, x.is_updatable] == [True, True, True]



from salesforce.testrunner.settings import *  # NOQA pylint: disable=unused-wildcard-import,wildcard-import
from salesforce.testrunner.settings import DATABASES, INSTALLED_APPS

SF_LAZY_CONNECT = True
# backup the static setting and set an empty dynamic setting
DATABASES = {
    'default': DATABASES['default'],

    'salesforce_copy': DATABASES['salesforce'],

    'salesforce': {
        'ENGINE': 'salesforce.backend',
        'AUTH': 'salesforce.auth.DynamicAuth',
    }
}
INSTALLED_APPS += ('tests.test_dynamic_auth',)



"""Backward compatible behaviour with primary key 'Id' and upper-case field names"""

import datetime
from salesforce import models
from salesforce.models import SalesforceModel


class User(SalesforceModel):
    Username = models.CharField(max_length=80)
    Email = models.CharField(max_length=100)


class Lead(SalesforceModel):
    Company = models.CharField(max_length=255)
    LastName = models.CharField(max_length=80)
    Owner = models.ForeignKey(User, on_delete=models.DO_NOTHING,
                              default=models.DEFAULTED_ON_CREATE, db_column='OwnerId')


# models for unit tests used without a connection only

class A(SalesforceModel):
    email = models.EmailField(custom=True)

    class Meta:
        db_table = 'A__c'


class B(SalesforceModel):

    class Meta:
        db_table = 'B__c'


class AtoB(SalesforceModel):
    a = models.ForeignKey(A, models.DO_NOTHING, custom=True)
    b = models.ForeignKey(B, models.DO_NOTHING, custom=True)

    class Meta:
        db_table = 'AtoB__c'


class TryDefaults(SalesforceModel):
    # this model doesn't exist in Salesforce, but it should be valid
    # it is only for coverage of code by tests
    example_str = models.CharField(max_length=50, default=models.DefaultedOnCreate('client'))
    example_datetime = models.DateTimeField(default=models.DefaultedOnCreate(datetime.datetime(2021, 3, 31, 23, 59)))
    # example_date = models.DateTimeField(default=models.DefaultedOnCreate(datetime.date(2021, 3, 31)))
    example_time = models.DateTimeField(default=models.DefaultedOnCreate(datetime.time(23, 59)))
    example_foreign_key = models.ForeignKey(User, on_delete=models.DO_NOTHING, default=models.DefaultedOnCreate())
    # ,default=models.DefaultedOnCreate(User(pk='000000000000000')))
    example_bool = models.BooleanField(default=models.DefaultedOnCreate(True))
    example_bool_2 = models.BooleanField(default=models.DefaultedOnCreate(False))



from salesforce.testrunner.settings import *  # NOQA pylint: disable=unused-wildcard-import,wildcard-import
from salesforce.testrunner.settings import INSTALLED_APPS

INSTALLED_APPS = [x for x in INSTALLED_APPS if x != 'salesforce.testrunner.example']
INSTALLED_APPS += ['tests.test_migrate']
DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': 'db_tmp_default',
    },
    'salesforce': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': 'db_tmp_salesforce',
    },
}
ROOT_URLCONF = None



# The tests of "dbapi/test_helpers.py" are here, because
# - they should not depend on Django
# - it is not nice to see '(expected failures=1)' in the main test results
#   if we really test the @expectedFailureIf decorator

from unittest import TestCase

from salesforce.dbapi import driver
from salesforce.dbapi.common import get_thread_connections
from salesforce.dbapi.test_helpers import LazyTestMixin, expectedFailureIf

# simulate that a connection exists. It is necessary for `lazy_check()`.
get_thread_connections()['dummy'] = 'dummy'


class TestExpectedFailure(TestCase):

    @expectedFailureIf(False)
    def test_condition_false(self):
        assert True

    @expectedFailureIf(True)
    def test_condition_true(self):
        assert False


class TestLazyAssert(LazyTestMixin, TestCase):

    def test_lazy_assert_fail(self):
        """example of an failed test if any lazy_assert fails"""
        self.lazyAssertTrue(True)
        self.lazyAssertTrue(False)
        self.lazyAssertTrue(True)
        # here would pass some important test
        with self.assertRaises(self.failureException):
            self.lazy_check()

    def test_error(self):
        """example that an error is more important than a failed lazy_assert"""
        self.lazyAssertTrue(False)
        with self.assertRaises(ZeroDivisionError):
            1 / 0
        with self.assertRaises(self.LazyAssertionError):
            self.lazy_check()

    def test_ok(self):
        """example that more asserts can be checked together"""
        self.lazyAssertTrue(True)
        self.lazyAssertTrue(True)
        self.lazy_check()

    def test_finally(self):
        """example of a test with a cean up after "finally:"

        The command "lazy_check()" should be before "finally:", because:
          - lazy_check() can raise probably.
          - only the more important exception should be raised if it occurs
        """
        with self.assertRaises(self.failureException):
            try:
                self.lazyAssertTrue(False)
                # other code that can fail
                # ...
                self.lazy_check()
            finally:
                clean_up = True  # some cleanup
        self.assertTrue(clean_up)


class TestLazyAssertNoSetup(LazyTestMixin, TestCase):
    """Verify that LazyTestMixin works also if omitted super().setUp()"""

    def setUp(self):
        # verify that super().setUp() is not necessary for LazyTestMixin
        pass

    def test_fail(self):
        self.lazyAssertTrue(False)
        with self.assertRaises(self.failureException):
            self.lazy_check()

    def test_ok(self):
        self.lazyAssertTrue(True)
        self.lazy_check()


class LazyAssertRequests(LazyTestMixin, TestCase):

    def test_ok(self):
        """example that more asserts can be checked together"""
        try:
            with self.lazy_assert_n_requests(1):
                driver.request_count += 1
            self.lazy_check()
        finally:
            pass  # some cleanup

    def test_fail(self):
        """example of an failed test of requests"""
        with self.lazy_assert_n_requests(1):
            driver.request_count += 2
        with self.assertRaises(self.failureException):
            self.lazy_check()

    def test_error(self):
        """example that a more important error can be caught later
        or than an assert in a lazy test"""
        with self.lazy_assert_n_requests(1):
            driver.request_count += 2  # then check requests
        with self.assertRaises(ZeroDivisionError):
            1 / 0
        with self.assertRaises(self.LazyAssertionError) as cm:
            self.lazy_check()
        exc_class, exc_value, exc_traceback_str = cm.test_case.lazy_failure
        self.assertTrue(issubclass(exc_class, AssertionError))
        self.assertEqual(
            ''.join(exc_value.args),
            '1 != 2 : expected requests != real requests;  checked by:\n    with self.lazy_assert_n_requests(1):'
        )
        self.assertRegex(exc_traceback_str, r'^File ".*", line ')



from salesforce.testrunner.settings import *  # NOQA pylint: disable=unused-wildcard-import,wildcard-import
from salesforce.testrunner.settings import INSTALLED_APPS

# replace the test app
INSTALLED_APPS = [x for x in INSTALLED_APPS if not x.startswith('salesforce.testrunner.')]
INSTALLED_APPS += ['tests.inspectdb']
ROOT_URLCONF = 'tests.inspectdb.urls'



"""Backward compatible behaviour with primary key 'Id' and upper-case field names"""

from salesforce import models
from salesforce.models import SalesforceModel


class User(SalesforceModel):
    username = models.CharField(max_length=80)
    email = models.CharField(max_length=100)
    last_name = models.CharField(max_length=80)
    first_name = models.CharField(max_length=40, null=True, blank=True)
    is_active = models.BooleanField(default=False)


class Lead(SalesforceModel):
    company = models.CharField(max_length=255)
    last_name = models.CharField(max_length=80)

    class Meta:
        db_table = 'Lead'


class Contact(SalesforceModel):
    last_name = models.CharField(max_length=80)
    owner = models.ForeignKey(User, on_delete=models.DO_NOTHING, default=models.DefaultedOnCreate(User))

    class Meta:
        managed = True
        db_table = 'Contact'



from django.test import TestCase
from django.conf import settings
from django.db import connections
from salesforce.testrunner.example.models import User
from requests.exceptions import ConnectionError


class LazyTest(TestCase):
    databases = '__all__'

    def test_lazy_connection(self):
        """
        Verify that the plain access to SF connection object does not raise
        exceptions with SF_LAZY_CONNECT if SF is not accessible.
        """
        # verify that access to a broken connection does not raise exception
        sf_conn = connections['salesforce']
        # try to authenticate on a temporary broken host
        users = User.objects.all()
        with self.assertRaises(Exception) as cm:
            len(users[:5])
        exc = cm.exception
        self.assertTrue(isinstance(exc, (ConnectionError, LookupError)))
        # fix the host name and verify that the connection works now
        sf_conn.connection.settings_dict.update(settings.ORIG_SALESFORCE_DB)
        self.assertGreater(len(users[:5]), 0)



# django-salesforce
#
# by Phil Christensen
# (c) 2012-2013 Freelancers Union (http://www.freelancersunion.org)
# See LICENSE.md for details
#

"""
This file demonstrates writing tests using the unittest module. These will pass
when you run "manage.py test".

Replace this with more appropriate tests for your application.
"""

# from django.test import TestCase



"""Recorded tests by MockTestCase"""
from typing import cast
import datetime
import decimal
import unittest
import django
from django.db import connections
from django.test.utils import override_settings

from .models import Account, Contact, Unreal
from tests.test_mock.mocksf import MockJsonRequest, MockRequest, MockTestCase  # mock
from salesforce.backend import DJANGO_30_PLUS
from salesforce.dbapi.exceptions import SalesforceError


@override_settings(SF_MOCK_MODE='mixed')
class TestMock(MockTestCase):
    api_version = '51.0'
    databases = {'salesforce'}

    @override_settings(SF_MOCK_MODE='record')
    @override_settings(SF_MOCK_VERBOSITY=0)
    def test_simple_crud(self) -> None:
        """Record the commands to a variable (quiet verbosity)"""
        acc = cast(Account, Account.objects.create(name='a'))
        acc = Account.objects.get(id=acc.id)
        acc.name = 'b'
        acc.save()
        acc.delete()
        output = django.db.connections['salesforce'].connection._sf_session.mock_recorded
        self.assertIn('POST mock:///services/data/', output[0])
        self.assertIn('GET mock:///services/data/', output[1])
        self.assertIn('PATCH mock:///services/data/', output[2])
        self.assertIn('DELETE mock:///services/data/', output[3])

    @override_settings(SF_MOCK_MODE='playback')
    def test_simple_crud_replay(self) -> None:
        """Replay the recorded commands and verify that also requests are exactly repeated"""
        optional_sql = '+LIMIT+21' if DJANGO_30_PLUS else ''
        self.mock_add_expected([
            MockJsonRequest(
                "POST mock:///services/data/v51.0/sobjects/Account",
                req={'Name': 'a'},
                resp='{"id":"001M000001FgVKlIAN","success":true,"errors":[]}',
                status_code=201),
            MockJsonRequest(
                "GET mock:///services/data/v51.0/query/?q=SELECT+Account.Id%2C+Account.Name%2C+Account.OwnerId+"
                "FROM+Account+WHERE+Account.Id+%3D+%27001M000001FgVKlIAN%27" + optional_sql,
                resp='{"totalSize":1,"done":true,"records":['
                '{"attributes":{"type":"Account","url":"/services/data/v51.0/sobjects/Account/001M000001FgVKlIAN"},'
                '"Id":"001M000001FgVKlIAN","Name":"a","OwnerId":"005M0000007whduIAA"}]}'),
            MockJsonRequest(
                "PATCH mock:///services/data/v51.0/sobjects/Account/001M000001FgVKlIAN",
                req={'Name': 'b', 'OwnerId': '005M0000007whduIAA'},
                request_type='',
                status_code=204),
            MockRequest(
                "DELETE mock:///services/data/v51.0/sobjects/Account/001M000001FgVKlIAN",
                request_type='',
                status_code=204)
        ])

        acc = cast(Account, Account.objects.create(name='a'))
        acc = Account.objects.get(id=acc.id)
        acc.name = 'b'
        acc.save()
        acc.delete()

    @override_settings(SF_MOCK_MODE='playback')
    def test_defaulted_on_create(self) -> None:
        """Replay different types of data"""
        self.mock_add_expected(MockJsonRequest(
            "POST mock:///services/data/v51.0/sobjects/Unreal",
            {'IntX': '1', 'DecimalX': 3.333, 'DatetimeX': '2021-04-14T18:52:45.000+0000',
             'StrX': 'a', 'Normal': '', 'Str2X': 'b', 'DateX': '2021-04-14',
             'TimeX': '11:59:59.000000Z', 'FloatX': '3.14', 'BoolX': 'true'
             },
            resp='[{"errorCode":"NOT_FOUND","message":"The requested resource does not exist"}]',
            status_code=404)
        )
        with self.assertRaises(SalesforceError) as exc:
            _ = Unreal.objects.create(
                bool_x=True,
                int_x=1,
                float_x=3.14,
                decimal_x=decimal.Decimal('3.333'),
                str_x='a',
                date_x=datetime.date(2021, 4, 14),
                datetime_x=datetime.datetime(2021, 4, 14, 18, 52, 45).replace(tzinfo=datetime.timezone.utc),
                time_x=datetime.time(11, 59, 59),
                str2_x='b',
            )
        self.assertEqual(exc.exception.data,
                         [{'errorCode': 'NOT_FOUND', 'message': 'The requested resource does not exist'}])

    @override_settings(SF_MOCK_MODE='playback')
    def test_defaulted_on_create_implicit(self) -> None:
        """Replay an example that ends with an error"""
        self.mock_add_expected(MockJsonRequest(
            "POST mock:///services/data/v51.0/sobjects/Unreal",
            {'Normal': ''},
            resp='{"id":"003000000000123AAA","success":true,"errors":[]}',
            status_code=201)
        )
        obj = Unreal()
        obj.save()
        self.assertEqual(obj.pk, "003000000000123AAA")


class Test(unittest.TestCase):
    databases = {'salesforce'}

    def is_complete(self) -> bool:
        cur = connections['salesforce'].cursor()
        cur.execute("select QualifiedApiName from FieldDefinition "
                    "where EntityDefinition.QualifiedApiName = 'Contact' "
                    "and QualifiedApiName='Donor_class__c'")
        return bool(len(cur.fetchall()))

    def test(self) -> None:
        if not self.is_complete():
            self.skipTest("Not found custom field 'Donor_class__c'")

        obj = Contact(last_name='a')
        obj.save()
        obj = Contact.objects.get(id=obj.id)
        # obj._state.__dict__.update({'db': 'salesforce', 'adding': False})
        obj.donor_class = None  # None can be saved to Salesforce, even if null=False
        obj.save()
        obj.delete()



from salesforce.testrunner.settings import *  # NOQA
from salesforce.testrunner.settings import INSTALLED_APPS

INSTALLED_APPS = [x for x in INSTALLED_APPS if not x.startswith('salesforce.testrunner.')]
# INSTALLED_APPS += ('tests.inspectdb', 'tests.inspectdb.dependent_model')
INSTALLED_APPS += ['tests.inspectdb.dependent_model.AutoModelConf',
                   'tests.inspectdb.dependent_model.DependentModelConf']
ROOT_URLCONF = None



import re
import sys
from django.test import TestCase
import django


class VersionTest(TestCase):
    def test_version(self):
        """Check that the expected Django version is being tested

        "tox" sometimes installed incorrect Django versions installed by "tox",
        e.g. if versions required by tox deps and requirements.txt or setup.py
        are in conflict. The righ version could be immediately uninstalled and
        replaced by an invalid version.
        """
        match = re.search(r'/\.tox/dj(\d+)-py(\d+)/', django.__file__)
        if match:
            django_version_abbr, python_version_abbr = match.groups()
            self.assertEqual('{}{}'.format(*django.VERSION[:2]), django_version_abbr)
            self.assertEqual('{}{}'.format(*sys.version_info[:2]), python_version_abbr)



import unittest
import requests
from salesforce import auth
from salesforce.dbapi import driver
import salesforce.testrunner.settings

sf_alias = 'salesforce'
settings_dict = salesforce.testrunner.settings.DATABASES[sf_alias]


class Test(unittest.TestCase):

    def test_no_django(self):
        self.assertRaises(ImportError, __import__, 'django.core')

    def test_auth_standard(self):
        auth.SalesforcePasswordAuth(sf_alias, settings_dict=settings_dict)


class OAuthTest(unittest.TestCase):

    def validate_oauth(self, settings_dict):
        for key in ('access_token', 'id', 'instance_url', 'issued_at', 'signature'):
            if key not in settings_dict:
                self.fail("Missing %s key in returned oauth data." % key)
            elif not settings_dict[key]:
                self.fail("Empty value for %s key in returned oauth data." % key)

    def test_token_renewal(self):
        # _session=salesforce.backend.fake.base.FakeAuthSession()
        # _session.bind('default')
        _session = requests.Session()

        auth_obj = auth.SalesforcePasswordAuth(sf_alias, settings_dict=settings_dict,
                                               _session=_session)
        auth_obj.get_auth()
        self.validate_oauth(auth.oauth_data[sf_alias])
        old_data = auth.oauth_data

        self.assertIn(sf_alias, auth.oauth_data)
        auth_obj.del_token()
        self.assertNotIn(sf_alias, auth.oauth_data)

        _session.close()  # close to prevent ResourceWarning: unclosed <ssl.SSLSocket...>
        auth_obj.get_auth()
        self.validate_oauth(auth.oauth_data[sf_alias])

        self.assertEqual(old_data[sf_alias]['access_token'], auth.oauth_data[sf_alias]['access_token'])
        _session.close()


class CursorTest(unittest.TestCase):
    """Currently tested only methods not covered by Django test"""

    def test_scroll(self):
        with driver.get_connection('salesforce', settings_dict=settings_dict) as connection:
            # the context manager for cursor is not important
            with connection.cursor() as cursor:
                cursor.execute("select Name from Contact limit 5")
                result = cursor.fetchall()
                # no more data
                self.assertEqual(cursor.fetchone(), None)
                self.assertEqual(cursor.fetchall(), [])
                self.assertEqual(cursor.fetchmany(), [])
                self.assertEqual(cursor.fetchmany(size=10), [])
                # scroll to replay the data
                cursor.scroll(0, 'absolute')
                result2 = cursor.fetchall()
                self.assertEqual(result2, result)



from salesforce import models

from tests.inspectdb.dependent_model import models_template


class User(models.Model):
    Username = models.CharField(max_length=80)


class Organization(models.Model):
    # all fields are copied dynamically
    class Meta:
        db_table = 'Organization'
        dynamic_field_patterns = models_template, ['.*']



"""Mock requests for Salesforce (REST/SOAP API to database)

Principial differences to other packages:
- The same query should have different results before and after insert, update, delete
  (therefore not used "requests-mock" etc.)
- This module has several modes:
  "record": mode for re-writing small integration tests to create mock tests
  "playback": mode for running the tests fast from recorded data
  "mixed" mode is used like a silent "record" mode. "mixed" is prepared to be
      switched to "record" mode inside the same session, e.g if a test test is
      extended by additional requests

    properties of modes:
        two sources or requests: application / recorded
        two sources or responses: Force.com server / recorded

        need to authentize before session?: (bool)
        record raw data or anonimize all ID
        record the traffic or to check and terminate at the first difference or
            to translate the historyrepla?
        cleanup after error or wait?
        a difference should be reported, but it an exception should not be raised before tearDown?
        * from application to server: * like a normal test
                                      * record all
                                      * check
                                      * check and record after difference
        * from playback to server: (check that the Force.com API is not changed)
        * nothing to server: response from playback: * only report differences
                                                     * stop a different request
                                                     * switch to server (replay and translate history)
        compare recorded requests
        compare test data

        Use raw recorded traffic or to try to find a nice formated equivalent record?
Parameters
    request_type: (None, 'application/json',... '*') The type '*' is for
        requests where the type and data should not be checked.
    data:  The recommended types are (str, None) A "dict" is possible, but
        it is not explicit enough for some data types.
    json:  it is unused and will be probably deprecated.
           It can be replaced by "json.dumps(request_json)"
"""
from typing import Any, Callable, cast, Dict, Iterable, Iterator, List, Optional, Set, Tuple, Type, Union
from unittest import mock, TestCase  # pylint:disable=unused-import  # NOQA
import json as json_mod
import re

import requests.models
from django.db import connections
from django.test import SimpleTestCase

from django.conf import settings

import salesforce
# from salesforce.dbapi import settings
from salesforce.auth import MockAuth, SalesforceAuth
from salesforce.dbapi import driver
from salesforce.dbapi.common import TimeStatistics
from salesforce.dbapi.driver import SfSession
from salesforce.dbapi.exceptions import DatabaseError
from salesforce.backend import DJANGO_22_PLUS
from salesforce.backend.test_helpers import sf_alias

AnyResponse = Union[requests.models.Response, 'MockResponse']

APPLICATION_JSON = 'application/json;charset=UTF-8'


dummy_login_session = SfSession()

# the first part are not test cases, but helpers for a mocked network: MockTestCase, MockRequest


class MockRequestsSession:
    """Prepare mock session with expected requests + responses history

    expected:   iterable of MockJsonRequest
    testcase:  testcase object (for consistent assertion)
    """

    def __init__(self, testcase: SimpleTestCase, expected: Iterable['MockRequest'] = (),
                 auth: Optional[SalesforceAuth] = None, old_session: Optional[SfSession] = None) -> None:
        self.index = 0  # index to self.expected
        self.testcase = testcase
        self.expected = list(expected)
        self.auth = auth or MockAuth('dummy alias', {'USER': ''}, _session=dummy_login_session)
        self.old_session = old_session
        self.mock_recorded = []  # type: List[str]

    def add_expected(self, expected_requests: Union['MockRequest', Iterable['MockRequest']]) -> None:
        if not isinstance(expected_requests, MockRequest):
            self.expected.extend(expected_requests)
        else:
            self.expected.append(expected_requests)

    def request(self, method: str, url: str, data: Optional[str] = None, **kwargs: Any) -> AnyResponse:
        """Assert the request equals the expected, return a historical response"""
        # pylint:disable=too-many-locals
        mode = getattr(settings, 'SF_MOCK_MODE', 'playback')
        if mode == 'playback':
            expected = self.expected[self.index]
            msg = "Difference at request index %d (from %d)" % (self.index, len(self.expected))
            response = expected.request(method, url, data=data, testcase=self.testcase,
                                        msg=msg, **kwargs)  # type: AnyResponse
            self.index += 1
            return response
        if mode in ('record', 'mixed'):
            if not self.old_session:
                raise ValueError(
                    'If set SF_MOCK_MODE="record" then the global value or value '
                    'in setUp method must be "mixed" or "record".')
            new_url = url.replace('mock://', self.old_session.auth.instance_url)
            response = self.old_session.request(method, new_url, data=data, **kwargs)
            if mode == 'record':
                output = []
                output.append('"%s %s"' % (method, url))
                if data:
                    output.append("req=%r" % data)
                if 'json' in kwargs:
                    output.append("request_json=%r" % kwargs['json'])
                if response.text:
                    output.append("resp=%r" % response.text)

                request_type = kwargs.get('headers', {}).get('Content-Type', '')
                response_type = response.headers.get('Content-Type', '')
                basic_type = request_type or response_type
                if basic_type.startswith('application/json'):
                    request_class = MockJsonRequest  # type: Type[MockRequest]
                else:
                    request_class = MockRequest
                    output.append("request_type=%r" % request_type)

                if response_type and (response_type != basic_type or request_class is MockRequest):
                    output.append("response_type=%r" % response_type)
                if response.status_code != 200:
                    output.append("status_code=%d" % response.status_code)
                new_output = (
                    "=== MOCK RECORD {testcase}\n{class_name}(\n    {params})\n===".format(
                        class_name=request_class.__name__,
                        testcase=self.testcase,
                        params=',\n    '.join(output))
                )
                self.mock_recorded.append(new_output)
                if self.verbosity > 0:
                    print(new_output, '\n')
            return response
        raise NotImplementedError("Not implemented SF_MOCK_MODE=%s" % mode)

    def get(self, url: str, **kwargs: Any) -> AnyResponse:
        return self.request('GET', url, **kwargs)

    def post(self, url: str, data: Optional[str] = None, json: Any = None, **kwargs: Any) -> AnyResponse:  # pylint:disable=redefined-outer-name # noqa
        return self.request('POST', url, data=data, json=json, **kwargs)

    def patch(self, url: str, data: Optional[str] = None, json: Any = None, **kwargs: Any) -> AnyResponse:  # pylint:disable=redefined-outer-name # noqa
        return self.request('PATCH', url, data=data, json=json, **kwargs)

    def delete(self, url: str, **kwargs: Any) -> AnyResponse:
        return self.request('DELETE', url, **kwargs)

    def mount(self, prefix: str, adapter: requests.adapters.HTTPAdapter) -> None:
        pass

    def close(self) -> None:
        pass

    @property
    def verbosity(self) -> int:
        return cast(int, getattr(settings, 'SF_MOCK_VERBOSITY', 1))


class MockRequest:
    """Recorded Mock request to be compared and response to be used

    for some unit tests offline
    If the parameter 'request_type' is '*' then the request is not tested
    """
    # pylint:disable=too-many-instance-attributes
    default_type = None  # type: Optional[str]

    def __init__(self, method_url: str,
                 req: Union[str, Dict[str, Any], None] = None, resp: Optional[str] = None,
                 request_json: Any = None,
                 request_type: Optional[str] = None, response_type: Optional[str] = None,
                 status_code: int = 200, check_request: bool = True) -> None:
        method, url = method_url.split(' ', 1)
        self.method = method
        self.url = url
        self.request_data = req
        self.response_data = resp
        self.request_json = request_json
        self.request_type = request_type or (self.default_type if method not in ('GET', 'DELETE') else '') or ''
        self.response_type = response_type
        self.status_code = status_code
        self.check_request = check_request

    def request(self, method: str, url: str, data: Optional[str] = None, json: Any = None,
                testcase: Optional[SimpleTestCase] = None, **kwargs: Any) -> 'MockResponse':
        # pylint:disable=too-many-branches
        """Compare the request to the expected. Return the expected response.
        Supported kwargs: 'msg', 'headers', 'timeout', 'verify'
        """
        if testcase is None:
            raise TypeError("Required keyword argument 'testcase' not found")
        msg = kwargs.pop('msg', None)
        if self.check_request:
            testcase.assertEqual(method.upper(), self.method.upper())
            testcase.assertEqual(url, self.url, msg=msg)
        if self.response_data:
            response_class = MockJsonResponse if self.default_type == APPLICATION_JSON else MockResponse
        else:
            response_class = MockResponse
        request_type = ''
        if json:
            assert not data
            data = json_mod.dumps(json)
            request_type = APPLICATION_JSON
        if (data or json) and 'headers' in kwargs:
            request_type = kwargs['headers'].pop('Content-Type', '') or request_type
        response = response_class(self.response_data,
                                  status_code=self.status_code,
                                  resp_content_type=self.response_type)
        response.request = RequestHint(method, url, body=data, headers={'content-type': request_type})
        if not self.check_request:
            return response

        if 'json' in self.request_type:
            assert data is not None and self.request_data is not None
            testcase.assertJSONEqual(data, self.request_data, msg=msg)
        elif 'xml' in self.request_type:
            assert data is not None and isinstance(self.request_data, str)
            testcase.assertXMLEqual(data, self.request_data, msg=msg)
        elif self.request_type != '*':
            testcase.assertEqual(data, self.request_data, msg=msg)
        if self.request_type != '*':
            request_json = kwargs.pop('json', None)
            testcase.assertEqual(request_json, self.request_json, msg=msg)
        if self.request_type != '*':
            testcase.assertEqual(request_type.split(';')[0], self.request_type.split(';')[0], msg=msg)
        kwargs.pop('timeout', None)
        assert kwargs.pop('verify', True) is True  # TLS verify must not be False
        if 'headers' in kwargs and not kwargs['headers']:
            del kwargs['headers']
        if kwargs:
            raise NotImplementedError("unexpected args = %s (msg=%s) at url %s" % (kwargs, msg, url))
        return response


class RequestHint:
    """Minimalist imitation of requests.models.Request usable for MockResponse"""
    def __init__(self, method: str, url: str, body: Optional[str], headers: Dict[str, str]) -> None:
        self.method = method
        self.url = url
        self.body = body
        self.headers = headers


class MockJsonRequest(MockRequest):
    """Mock JSON request/response for some unit tests offline"""
    default_type = APPLICATION_JSON


class MockTestCase(SimpleTestCase):
    """
    Test case that uses recorded requests/responses instead of network
    """
    if DJANGO_22_PLUS:
        databases = {'salesforce'}  # type: Union[Set[str], str]
    else:
        allow_database_queries = True

    def setUp(self) -> None:
        # pylint:disable=protected-access
        mode = getattr(settings, 'SF_MOCK_MODE', 'playback')
        super(MockTestCase, self).setUp()
        connection = connections[sf_alias]
        if connection.vendor != 'salesforce':
            raise DatabaseError("MockTestCase can run only on Salesforce databases")
        if not connection.connection:
            connection.connect()
        if mode != 'playback':
            # if the mode is 'record' or 'mixed' we must create a real connection before mock
            if not connection.connection.sf_session:
                connection.make_session()
        connection = connection.connection
        self.sf_connection = connection
        self.save_session_auth = connection._sf_session, connection.sf_auth
        connection._sf_session = MockRequestsSession(testcase=self, old_session=connection._sf_session)
        connection.sf_auth = connection._sf_session.auth

        self.save_api_version = connection._api_version
        connection._api_version = getattr(self, 'api_version', salesforce.API_VERSION)
        # simulate a recent request (to not run a check for broken conection in the test)
        driver.time_statistics.expiration = 1E10

    def tearDown(self) -> None:
        result = self.defaultTestResult()  # these 2 methods have no side effects
        self._feedErrorsToResult(result, self._outcome.errors)  # type: ignore[attr-defined] # hidden attributes used
        error = self.list2reason(result.errors)
        failure = self.list2reason(result.failures)
        ok = not error and not failure
        # # demo:   report short info immediately (not important)
        # if not ok:
        #     typ, text = ('ERROR', error) if error else ('FAIL', failure)
        #     msg = [x for x in text.split('\n')[1:] if not x.startswith(' ')][0]
        #     print("\n%s: %s\n     %s" % (typ, self.id(), msg))

        connection = self.sf_connection
        session = connection._sf_session  # pylint:disable=protected-access
        try:
            if ok and isinstance(session, MockRequestsSession):
                self.assertEqual(session.index, len(session.expected), "Not all expected requests has been used")
        finally:
            connection._sf_session, connection.sf_auth = self.save_session_auth  # pylint:disable=protected-access
            connection._api_version = self.save_api_version
            driver.time_statistics.expiration = TimeStatistics().expiration
            super(MockTestCase, self).tearDown()

    def list2reason(self, exc_list: List[Tuple[TestCase, str]]) -> Optional[str]:
        if exc_list and exc_list[-1][0] is self:
            return exc_list[-1][1]
        return None

    def mock_add_expected(self, expected_requests: Union[MockRequest, Iterable[MockRequest]]) -> None:
        self.sf_connection._sf_session.add_expected(expected_requests)  # pylint:disable=protected-access


# class MockXmlRequest - only with different default content types


class MockResponse:
    """Mock response for some unit tests offline"""
    default_type = None  # type: Optional[str]
    request = None  # type: RequestHint

    def __init__(self, text: Optional[str], resp_content_type: Optional[str] = None, status_code: int = 200) -> None:
        self.text = text
        self.status_code = status_code
        self.content_type = resp_content_type if resp_content_type is not None else self.default_type

    def json(self, parse_float: Optional[Callable[[str], Any]] = None) -> Any:
        assert self.text
        return json_mod.loads(self.text.replace('...', ''), parse_float=parse_float)

    @property
    def headers(self) -> Dict[str, str]:
        return {'Content-Type': self.content_type} if self.content_type else {}


class MockJsonResponse(MockResponse):
    default_type = APPLICATION_JSON


# Undocumented - useful for tests


def case_safe_sf_id(id_15: Optional[str]) -> Optional[str]:
    """
    Equivalent to Salesforce CASESAFEID()

    Convert a 15 char case-sensitive Id to 18 char case-insensitive Salesforce Id
    or check the long 18 char ID.

    Long  18 char Id are from SFDC API and from Apex. They are recommended by SF.
    Short 15 char Id are from SFDC formulas if omitted to use func CASESAFEID(),
    from reports or from parsed URLs in HTML.
    The long and short form are interchangable as the input to Salesforce API or
    to django-salesforce. They only need to be someway normalized if they are
    used as dictionary keys in a Python application code.
    """
    if not id_15:
        return None
    if len(id_15) not in (15, 18):
        raise TypeError("The string %r is not a valid Force.com ID")
    suffix = []
    for i in range(0, 15, 5):
        weight = 1
        digit = 0
        for char in id_15[i:i + 5]:
            if char not in '0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz':
                raise TypeError("The string %r is not a valid Force.com ID")
            if char.isupper():
                digit += weight
            weight *= 2
        suffix.append(chr(ord('A') + digit) if digit < 26 else str(digit - 26))
    out = ''.join(suffix)
    if len(id_15) == 18 and out != id_15[15:]:
        raise TypeError("The string %r is not a valid Force.com ID")
    return id_15[:15] + out


def check_sf_api_id(id_18: str) -> Optional[str]:
    """
    Check the 18 characters long API ID, no exceptions
    """
    try:
        return case_safe_sf_id(id_18)
    except TypeError:
        return None


def extract_ids(data_text: str, data_type: Optional[str] = None) -> Iterator[Tuple[str, Tuple[int, int]]]:
    """
    Extract all Force.com ID from REST/SOAP/SOQL request/response (for mock tests)

    Output: iterable of all ID and their positions
    Parameters: data_type:  can be in ('rest', 'soap', 'soql', None),
                    where None is for any unknown type
    """
    id_pattern = r'([0-9A-Za-z]{18})'
    pattern_map = {None: r'[">\']{}["<\']',
                   'rest': '"{}"',
                   'soap': '>{}<',
                   'soql': "'{}'"
                   }
    pattern = pattern_map[data_type].format(id_pattern)
    for match in re.finditer(pattern, data_text):
        txt = match.group(1)
        if case_safe_sf_id(txt):
            yield txt, (match.start(1), match.end(1))



# django-salesforce
#
# by Phil Christensen
# (c) 2012-2013 Freelancers Union (http://www.freelancersunion.org)
# See LICENSE.md for details
#
"""
DatabaseOperations  (like salesforce.db.backends.*.operations)
"""
from typing import Optional
import itertools
import warnings

import django.db.backends.utils
from django.db.backends.base.operations import BaseDatabaseOperations
from salesforce.backend import DJANGO_30_PLUS
from salesforce.dbapi.exceptions import SalesforceWarning

BULK_BATCH_SIZE = 200

"""
Default database operations, with unquoted names.
"""


class DatabaseOperations(BaseDatabaseOperations):  # pylint:disable=too-many-public-methods
    # undefined abstract methods:
    #       date_extract_sql, date_interval_sql,     date_trunc_sql,  datetime_cast_date_sql
    #   datetime_extract_sql,                    datetime_trunc_sql,
    #                                                time_trunc_sql,  datetime_cast_time_sql;
    #   no_limit_value,   regex_lookup
    #
    # pylint:disable=abstract-method,no-self-use,unused-argument

    compiler_module = "salesforce.backend.compiler"
    explain_prefix = 'EXPLAIN'

    def sql_flush(self, style, tables, *, reset_sequences=False, allow_cascade=False):
        raise NotImplementedError('Salesforce does not allow to flush all tables.')
        # return []

    def quote_name(self, name):
        return name

    if DJANGO_30_PLUS:

        def fetch_returned_insert_columns(self, cursor, returning_params=None):
            # the parameter "returning_params" is for Django 3.1
            return (cursor.lastrowid,)

        def fetch_returned_insert_rows(self, cursor):
            return [(x,) for x in cursor.lastrowid]

        def return_insert_columns(self, fields):
            return '', ()  # dummy result

    else:

        def fetch_returned_insert_id(self, cursor):
            return cursor.lastrowid

        def fetch_returned_insert_ids(self, cursor):
            return cursor.lastrowid

        def return_insert_id(self):
            return "", ()

    def max_in_list_size(self) -> Optional[int]:
        # A non-zero return value of max_in_list_size() should be never used
        # because it is useful only for Oracle and this form of splitting would
        # cause only problems in Salesforce and no benefit:
        #     WHERE (Id IN ('z0001',... 'z1000') OR ... (Id IN 'z5001',... 'z6000'))
        return None

    def max_name_length(self) -> Optional[int]:
        # this can not be implemented correctly because:
        # the maximum length of custom field is 40 and plus a suffix '__c' is added
        # plus a possible namespace prefix of max length 15 plus two undercores '__'.
        # The longest field name in some system table is 57
        return 60  # 15 + 2 + 40 + 3

    def adapt_datefield_value(self, value):
        return value

    def adapt_datetimefield_value(self, value):
        return value

    def adapt_timefield_value(self, value):
        return value

    def adapt_decimalfield_value(self, value, max_digits=None, decimal_places=None):
        if hasattr(value, 'default'):  # DefaultedOnCreate
            return value
        return django.db.backends.utils.format_number(value, max_digits, decimal_places)

    def bulk_batch_size(self, fields, objs):
        return BULK_BATCH_SIZE

    # This SQL is not important because we currently control the insert from a Salesforce compiler,
    # but some method must exist.
    def bulk_insert_sql(self, fields, placeholder_rows):
        return "VALUES " + ", ".join(itertools.chain(*placeholder_rows))

    def conditional_expression_supported_in_where_clause(self, expression):
        """
        All filter elements in a WHERE clause must be a field compared with value.
        The same is necessary for boolean fields, e.g. IsActive=true
        """
        return False

    def prep_for_like_query(self, x):
        """Prepare a value for use in a LIKE query."""
        if '\\' in x:
            warnings.warn("Backslash not allowed in LIKE expressions for Salesforce", SalesforceWarning)
        # A wildcard search is better than a search of '\\%' or '\\_', see #254
        return str(x)
        # return str(x).replace("\\", "\\\\").replace("%", r"\%").replace("_", r"\_")


def DefaultedOnCreate(*args, **kwargs):
    import salesforce.defaults  # pylint:disable=import-outside-toplevel
    warnings.warn("Deprecated: the object DefaultedOnCreate should be imported from "
                  "salesforce.fields or salesforce.defaults or salesforce.models, "
                  "but not from salesforce.backend.operations")
    return salesforce.defaults.DefaultedOnCreate(*args, **kwargs)



from django.db import connections

from salesforce.dbapi.exceptions import SalesforceError
from tests.test_mock.mocksf import MockJsonRequest, MockRequest, MockTestCase
from tests.test_mock.mocksf import mock  # NOQA pylint:disable=unused-import

# :%s@mock://@mock://@


class MockTest(MockTestCase):
    api_version = '20.0'

    def setUp(self) -> None:
        super(MockTest, self).setUp()
        self.cursor = connections['salesforce'].cursor()

    # ---------------------

    def test_version(self) -> None:
        "Get the Salesforce Version"
        self.mock_add_expected(
            MockJsonRequest(
                "GET mock:///services/data/",
                resp="""[
                    {
                        "version":"20.0",
                        "url":"/services/data/v20.0",
                        "label":"Winter '11"
                    }
                    ...
                ]""",))
        ret = self.cursor.cursor.versions_request()
        self.assertEqual(ret, [{'version': '20.0', 'url': '/services/data/v20.0', 'label': "Winter '11"}])

    # ---------------------

    def test_resources(self) -> None:
        "Get a List of Resources"
        self.mock_add_expected(
            MockJsonRequest(
                "GET mock:///services/data/v20.0/",
                resp="""{
                    "sobjects" : "/services/data/v20.0/sobjects",
                    "search" : "/services/data/v20.0/search",
                    "query" : "/services/data/v20.0/query",
                    "recent" : "/services/data/v20.0/recent"
                    ...
                }""",))
        ret = self.cursor.cursor.urls_request()
        self.assertEqual(ret['sobjects'], '/services/data/v20.0/sobjects')

    # ---------------------

    def test_foo(self) -> None:
        "..."
        self.mock_add_expected(
            MockJsonRequest(
                "GET mock:///services/data/v20.0/",
                resp="""
                {
                    ...
                }""",))
        ret = self.cursor.cursor.urls_request()
        self.assertEqual(ret, {})

# ---------------------


class SObjectCollectionsTest(MockTestCase):
    """
    Examples from SObject Collections
    https://developer.salesforce.com/docs/atlas.en-us.api_rest.meta/api_rest/resources_composite_sobjects_collections.htm
    """
    api_version = '42.0'

    def setUp(self) -> None:
        super(SObjectCollectionsTest, self).setUp()
        self.cursor = connections['salesforce'].cursor()  # TODO not important

    def test_create(self) -> None:
        "Create Multiple Records with Fewer Round-Trips"
        # This test data were recorded from REST API documentation in August 2016

        # Simulate a request and response call without all_or_none
        request_allornone_false = """
        {
           "allOrNone" : false,
           "records" : [{
              "attributes" : {"type" : "Account"},
              "Name" : "example.com",
              "BillingCity" : "San Francisco"
           }, {
              "attributes" : {"type" : "Contact"},
              "LastName" : "Johnson",
              "FirstName" : "Erica"
           }]
        }"""
        resp_1 = MockJsonRequest(
            "POST mock:///services/data/v42.0/composite/sobjects",
            request_allornone_false,
            resp="""
            [
               {
                  "id" : "001RM000003oLnnYAE",
                  "success" : true,
                  "errors" : [ ]
               },
               {
                  "id" : "003RM0000068xV6YAI",
                  "success" : true,
                  "errors" : [ ]
               }
            ]"""
        )
        self.mock_add_expected(resp_1)
        data = [
            dict(type_="Account", Name="example.com", BillingCity="San Francisco"),
            dict(type_="Contact", LastName="Johnson", FirstName="Erica")
        ]

        ret = self.cursor.db.connection.sobject_collections_request('POST', data, all_or_none=False)
        # the primary keys start with '001' (Account) and '003' (Contact'
        self.assertEqual([x[:3] for x in ret], ['001', '003'])

        # Simulate the same repeated request after the previous.
        # Expect that duplicates are not allowed by SFDC instance configuration
        resp_2 = MockJsonRequest(
            "POST mock:///services/data/v42.0/composite/sobjects",
            request_allornone_false,
            resp="""
            [
               {
                  "success" : false,
                  "errors" : [
                     {
                        "statusCode" : "DUPLICATES_DETECTED",
                        "message" : "Use one of these records?",
                        "fields" : [ ]
                     }
                  ]
               },
               {
                  "id" : "003RM0000068xVCYAY",
                  "success" : true,
                  "errors" : [ ]
               }
            ]""",
        )
        self.mock_add_expected(resp_2)
        with self.assertRaises(SalesforceError) as cm:
            ret = self.cursor.db.connection.sobject_collections_request('POST', data, all_or_none=False)
        self.assertIn('Account  DUPLICATES_DETECTED', cm.exception.args[0])

        # simulate the same, but with all_or_none. (One error reported and one operation rollback)
        request_allornone_true = request_allornone_false.replace('"allOrNone" : false', '"allOrNone" : true')
        resp_3 = MockJsonRequest(
            "POST mock:///services/data/v42.0/composite/sobjects",
            request_allornone_true,
            resp="""
            [
               {
                  "success" : false,
                  "errors" : [
                     {
                        "statusCode" : "DUPLICATES_DETECTED",
                        "message" : "Use one of these records?",
                        "fields" : [ ]
                     }
                  ]
               },
               {
                  "success" : false,
                  "errors" : [
                     {
                        "statusCode" : "ALL_OR_NONE_OPERATION_ROLLED_BACK",
                        "message" : "Record rolled back because not all records were valid and the request was """
                 """using AllOrNone header",
                        "fields" : [ ]
                     }
                  ]
               }
            ]""",
        )
        self.mock_add_expected(resp_3)
        with self.assertRaises(SalesforceError) as cm:
            ret = self.cursor.db.connection.sobject_collections_request('POST', data, all_or_none=True)
        self.assertIn('Account  DUPLICATES_DETECTED', cm.exception.args[0])

        # with mock.patch.object(self.cursor.db.connection, 'composite_type', 'composite'):
        #    ret = self.cursor.db.connection.sobject_collections_request('POST', data, all_or_none=True)


def parse_this() -> MockRequest:
    # OAuth error codes are in
    # https://support.salesforce.com/articleView?id=remoteaccess_errorcodes.htm&type=5

    # this is the error if the method was POST/PATCH and forgot Content-Type ".../json"
    #   response.text.startswith('<response><faultcode>LOGIN_OAUTH_INVALID_TOKEN')
    return MockRequest(
        '...', request_type='application/xml',
        status_code=401,
        # response headers {'Sfdc-OAuth-Error': '1703',...}
        response_type='text/html; charset=UTF-8',
        req="""
            <response>
            <faultcode>LOGIN_OAUTH_INVALID_TOKEN(db=1703,api=LOGIN_OAUTH_INVALID_TOKEN)</faultcode>
            <error>Failed: Invalid Token</error>
            </response>"""
    )



from django.apps import AppConfig


class AutoModelConf(AppConfig):
    name = 'tests.inspectdb'
    label = 'auto_model'


class DependentModelConf(AppConfig):
    name = 'tests.inspectdb.dependent_model'
    label = 'dependent_model'



"""Test that app config can override a directory name conflict (.e.g. "salesforce")"""
from salesforce.testrunner.settings import *  # NOQA pylint: disable=unused-wildcard-import,wildcard-import
from salesforce.testrunner.settings import INSTALLED_APPS

INSTALLED_APPS = [x for x in INSTALLED_APPS if x != 'salesforce.testrunner.example']
INSTALLED_APPS += ['tests.test_app_label.salesforce.apps.TestSalesForceConfig']
ROOT_URLCONF = None



from salesforce.testrunner.settings import *  # NOQA pylint: disable=unused-wildcard-import,wildcard-import
from salesforce.testrunner.settings import DATABASES

INSTALLED_APPS = ['tests.test_default_db']
DATABASES = {'default': DATABASES['default'], 'salesforce': DATABASES['salesforce']}
SALESFORCE_DB_ALIAS = 'default'
ROOT_URLCONF = None
# SF_LAZY_CONNECT = True



"""
Backward compatible behaviour with a primary key 'Id' and upper-case field names

"""

from django.db import models


class User(models.Model):
    username = models.CharField(max_length=80)
    last_name = models.CharField(max_length=80)
    first_name = models.CharField(max_length=40, null=True, blank=True)
    email = models.EmailField()
    is_active = models.BooleanField(default=False)


class Lead(models.Model):
    company = models.CharField(max_length=255)
    last_name = models.CharField(max_length=80)
    owner = models.ForeignKey(User, on_delete=models.DO_NOTHING)



"""Test djangoBackward compatible behaviour with primary key 'Id'."""
from django.test import TestCase

from .models import Account, Contact
from salesforce.backend.test_helpers import LazyTestMixin


class CompatibilityTest(TestCase, LazyTestMixin):
    databases = {'salesforce'}

    def test_alternate_default_manager(self) -> None:
        """Test an alternate default manager with a queryset"""
        test_acc = Account.objects.create(name='Aa')
        test_cnt = Contact.objects.create(last_name='B', account=test_acc)
        try:
            # test that only one request is used for .save()
            with self.lazy_assert_n_requests(1):
                test_cnt.save()

            # test again that only one request is used for .save() for a combined queryset
            wrk = Contact.objects.filter(last_name='B')
            tmp = wrk[0]
            tmp.last_name = 'C'
            with self.lazy_assert_n_requests(1):
                tmp.save()

            # check that the default queryset is not empty
            _ = Contact.objects.all()[0]

            with self.lazy_assert_n_requests(2):
                Contact.objects.sf(edge_updates=True).update(last_name='sf_test')

            self.lazy_check()
        finally:
            test_cnt.delete()
            test_acc.delete()



"""
Generated by 'django-admin startproject' using Django 3.0.5.
"""

import os
from typing import List
from django.utils.crypto import get_random_string

BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
SECRET_KEY = get_random_string(length=32)

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = True

ALLOWED_HOSTS = []  # type: List[str]

INSTALLED_APPS = [
    'tests.no_salesforce',
    # 'tests.test_default_db',
]
if os.environ.get('SF_WITH_EXAMPLE') == 'on':
    INSTALLED_APPS.append('salesforce.testrunner.example')

ROOT_URLCONF = None

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': os.path.join(BASE_DIR, 'db.sqlite3'),
    }
}


TIME_ZONE = 'UTC'
USE_TZ = True
STATIC_URL = '/static/'
DEFAULT_AUTO_FIELD = 'django.db.models.AutoField'
SALESFORCE_DB_ALIAS = 'default'
SF_EXAMPLE_EXTENDED_MODELS = True
SF_EXAMPLE_CUSTOM_INSTALLED = True



import debug_toolbar  # type: ignore[import] # noqa
from django.urls import include, path
from salesforce.testrunner.urls import urlpatterns
from .views import account_insert_delete

urlpatterns += [
    path('__debug__/', include(debug_toolbar.urls, namespace='djdt')),
    path('account_insert_delete/', account_insert_delete),
]



# django-salesforce
#
# by Phil Christensen
# (c) 2012-2013 Freelancers Union (http://www.freelancersunion.org)
# See LICENSE.md for details
#

"""
Default validation code.  (like django.db.backends.*.validation)
"""
from django.db.backends.base.validation import BaseDatabaseValidation


class DatabaseValidation(BaseDatabaseValidation):
    pass



"""Backward compatible behaviour with primary key 'Id' and upper-case field names"""

from salesforce import models
from salesforce.models import SalesforceModel
from salesforce.backend.manager import SalesforceManager
from salesforce.backend.query import SalesforceQuerySet


class ContactQuerySet(SalesforceQuerySet):
    def people(self):
        return self.filter(account__name__gt='A')


class FilteredManager(SalesforceManager):
    def get_queryset(self):
        return ContactQuerySet(self.model, using=self._db).people()

    def people(self):
        return self.get_queryset().people()


class Account(SalesforceModel):
    name = models.CharField(max_length=80)


class Contact(SalesforceModel):
    last_name = models.CharField(max_length=80)
    account = models.ForeignKey(Account, models.DO_NOTHING)
    objects = FilteredManager()



"""
Lookups  (like django.db.models.lookups, django.db.models.aggregates.Count)
"""
from django.db import models
from django.db.models.fields import Field
from django.db.models import lookups


class IsNull(models.lookups.IsNull):
    def override_as_sql(self, compiler, connection):  # pylint:disable=unused-argument
        # it must be relabeled if used for a children rows set
        if compiler.soql_trans is None:
            compiler.get_from_clause()
        sql, params = compiler.compile(self.lhs.relabeled_clone(compiler.soql_trans))
        return ('%s %s null' % (sql, ('=' if self.rhs else '!='))), params

    setattr(models.lookups.IsNull, 'as_salesforce', override_as_sql)


class Range(models.lookups.Range):
    def override_as_sql(self, compiler, connection):
        lhs, lhs_params = self.process_lhs(compiler, connection)
        rhs, rhs_params = self.process_rhs(compiler, connection)
        assert rhs == ('%s', '%s')
        assert len(rhs_params) == 2
        params = lhs_params + [rhs_params[0]] + lhs_params + [rhs_params[1]]
        # The symbolic parameters %s are again substituted by %s. The real
        # parameters will be passed finally directly to CursorWrapper.execute
        return '(%s >= %s AND %s <= %s)' % (lhs, rhs[0], lhs, rhs[1]), params

    setattr(models.lookups.Range, 'as_salesforce', override_as_sql)


class Count(models.aggregates.Count):
    # pylint:disable=abstract-method,too-many-ancestors  # undefined __and__, __or__, __rand__, __ror__

    def override_as_sql(self, *args, **kwargs):
        if len(self.source_expressions) == 1 and isinstance(self.source_expressions[0], models.expressions.Star):
            # patch for compile "Count('*')
            # (an alternative verbose solution is: f"COUNT({args[0].query.model._meta.db_table}.Id)")
            return 'COUNT(Id)', []
        # compile a normal Count('some_field')
        sql, params = self.as_sql(*args, **kwargs)
        if sql.startswith('COUNT(DISTINCT '):
            sql = sql.replace('COUNT(DISTINCT ', 'COUNT_DISTINCT(')
        return sql, params

    setattr(models.aggregates.Count, 'as_salesforce', override_as_sql)


@Field.register_lookup
class NotIn(lookups.In):
    lookup_name = 'not_in'

    def get_rhs_op(self, connection, rhs):
        return 'NOT IN %s' % rhs

    def split_parameter_list_as_sql(self, compiler, connection):
        # this patch is a only for Oracle and never tested with it
        max_size = connection.ops.max_in_list_size()
        raise NotImplementedError("Lookup 'not_in' can't be used with lists longer then %d" % max_size)


@Field.register_lookup
class NotEqual(lookups.Exact):
    lookup_name = 'not_eq'

    def get_rhs_op(self, connection, rhs):
        return '!= %s' % rhs



# django-salesforce
#
# by Phil Christensen
# (c) 2012-2013 Freelancers Union (http://www.freelancersunion.org)
# See LICENSE.md for details
#

"""
Multi-database support for the Django admin.
Not important yet - replaced by an empty stub.
"""
from django.contrib.admin.options import ModelAdmin as RoutedModelAdmin  # NOQA pylint:disable=unused-import



"""Test the completness ad validity of inspectdb by read and write for all tables.

It tries to find model and read one record from all retrieveable tables, except
objects with extraordinal filter conditions (3%). Then it tries to write this
record back if the table is updateable, except some tables (5%). This can fail
on permissions.

Usage:
$ python manage.py inspectdb --database=salesforce >tests/inspectdb/models.py
$ python manage.py check
$ python tests/inspectdb/slow_test.py
"""

from inspect import isclass
from sys import stdout, stderr
import os
import sys

import django
sys.path.insert(0, '.')
os.environ['DJANGO_SETTINGS_MODULE'] = 'tests.inspectdb.settings'

django.setup()

# these 3 lines must be imported after: path, environ, django.setup()
from django.db import connections  # NOQA
from tests.inspectdb import models as mdl  # NOQA
from salesforce.backend.base import SalesforceError  # NOQA


sf = connections['salesforce']


def run():
    start_name = sys.argv[1] if sys.argv[1:] else ''
    n_tables = n_read = n_no_data = n_read_errors = n_write = n_write_errors = 0
    model_map = {cls._meta.db_table: cls
                 for cls in mdl.__dict__.values()
                 if isclass(cls) and issubclass(cls, django.db.models.Model)
                 }
    problematic_read = {
        # These require specific filters (descried in their error messages)
        'CollaborationGroupRecord', 'ContentFolderMember', 'ContentFolderItem',
        'ContentDocumentLink', 'Idea', 'IdeaComment', 'UserProfileFeed',
        'Vote',  # 'OpportunityPartner', 'Product2Feed',
        # UNKNOWN_EXCEPTION:
        'TenantUsageEntitlement',
        # special
        'django_migrations',
        }
    problematic_write = {
        # Cannot modify managed objects
        'ApexClass', 'ApexComponent', 'ApexTrigger', 'FieldPermissions',
        'ObjectPermissions', 'PermissionSet', 'Scontrol',
        'StaticResource', 'WebLink', 'RecordType',
        # This is not writable due to 'NamespacePrefix' field
        'ApexPage',
        # It does not update the 'last_modified_date' field
        'AppMenuItem',
        # Some Leads are not writable becase they are coverted to Contact
        'Lead',
        # Insufficient access rights on cross-reference id
        'Group',
        'OpportunityShare', 'Profile',
        # Some very old items can have empty Folder.Name, but can
        # not be saved again without Name.
        'Folder',
        # Records with some values of UserShare.RowCause can not be updated.
        'UserShare',
        # Cannot directly insert FeedItem with type TrackedChange
        'FeedItem',
        # The Apex Class selected is not valid. An Apex Class that implements
        # the Messaging.InboundEmailHandler interface must be selected.
        'EmailServicesFunction',
        # A special user of type user_type='AutomatedProcess' exists in Users
        # since API 41.0 Winter'18.
        # That user account can not be saved because it has an invalid email.
        'User', 'UserLogin',
    }
    for tab in sf.introspection.table_list_cache['sobjects']:
        db_table = tab['name']
        # More sobjects are 'queryable' and 'retrieveable' than
        # only 'retrieveable', but half of them would be problematic.
        if tab['retrieveable'] and not db_table.endswith('ChangeEvent') and db_table not in problematic_read:
            if db_table < start_name:
                continue
            test_class = model_map[db_table]

            stdout.write('%s' % db_table)
            obj = None
            try:
                n_read += 1
                obj = test_class.objects.all()[0]
            except SalesforceError as exc:
                stderr.write("\n************** %s %s\n" % (db_table, exc))
                n_read_errors += 1
            except IndexError:
                n_no_data += 1
            if obj:
                stdout.write("* ")
            if obj and tab['updateable'] and db_table not in problematic_write:
                stdout.write('(write) ')
                try:
                    n_write += 1
                    obj.save(force_update=True)
                except SalesforceError as exc:
                    stderr.write("\n************** %s %s\n" % (db_table, exc))
                    n_write_errors += 1
                else:
                    # object 'Topic' doesn't have the attribute 'last_modified_date'
                    # in recently created SFDC databases (proably version 34.0+)
                    if hasattr(obj, 'last_modified_date'):
                        assert test_class.objects.get(pk=obj.pk).last_modified_date > obj.last_modified_date
            stdout.write('\n')
    n_tables = len(sf.introspection.table_list_cache['sobjects'])
    print('Inspectdb results:')
    print('  {n_tables} tables, {n_read} reads tried, {n_no_data} no data, '
          '{n_read_errors} read errors, {n_write} writes tried, {n_write_errors} write errors'
          .format(n_tables=n_tables, n_read=n_read, n_no_data=n_no_data,
                  n_read_errors=n_read_errors, n_write=n_write, n_write_errors=n_write_errors))
    print('********* ERRORs found' if n_read_errors + n_write_errors else 'OK')
    return n_read_errors + n_write_errors == 0


if __name__ == '__main__':
    ok = run()
    sys.exit(0 if ok else 1)



from salesforce.testrunner.settings import *  # NOQA pylint: disable=unused-wildcard-import,wildcard-import
from salesforce.testrunner.settings import INSTALLED_APPS, MIDDLEWARE

INSTALLED_APPS += ('debug_toolbar', 'tests.t_debug_toolbar',)
MIDDLEWARE = ['debug_toolbar.middleware.DebugToolbarMiddleware'] + MIDDLEWARE
ROOT_URLCONF = 'tests.t_debug_toolbar.urls'
INTERNAL_IPS = ['127.0.0.1']



"""
Clean test objects after an interrupted test.

All tests are written so that a failed test should usually not leave database objects,
but a test interrupted by debugger or Ctrl-C could do it.
"""
from typing import List
from unittest import TestCase
from django.db.models import Q
from salesforce.backend.operations import BULK_BATCH_SIZE
from salesforce.backend.query import SalesforceQuerySet
from salesforce.testrunner.example.models import Account, Campaign, Contact, Lead, Opportunity, Product, Test
from salesforce.tests.test_integration import sf_tables


def qs_delete(qs: SalesforceQuerySet) -> List[int]:
    """Delete all items of queryset"""
    ret = list(qs)
    s = 0
    for offs in range(0, len(ret), BULK_BATCH_SIZE):
        ids = [x.pk for x in ret[offs: offs + BULK_BATCH_SIZE]]
        s += qs.model.objects.filter(pk__in=ids).delete()[0]
    return [s]


class CleanTests(TestCase):

    def setUp(self):
        ret = [
            qs_delete(Account.objects.filter(
                Q(Name__startswith='sf_test') | Q(Name__startswith='test') |
                Q(Name__in=[r"""Dr. Evil's Giant\' "Laser", LLC""", "IntegrationTest Account"])
            )),
            qs_delete(Campaign.objects.filter(name__startswith='test ')),
            qs_delete(Contact.objects.filter(Q(first_name__startswith='sf_test') | Q(last_name__startswith='sf_test') |
                                             Q(name='Joe Freelancer'))),
            qs_delete(Lead.objects.filter(LastName__startswith='UnitTest')),
            qs_delete(Opportunity.objects.filter(name__in=('test op', 'Example Opportunity'))),
            qs_delete(Product.objects.filter(Name__startswith='test ')),
        ]
        if 'django_Test__c' in sf_tables():
            ret.append(qs_delete(Test.objects.all()))
        print("Deleted objects={}".format(sum(x[0] for x in ret)))

    def test_clean_all_test_data(self):
        """It is cleaninng all known test data that can be necessary sometimes after interrupted tests
        e.g. after debugging.
        It is not a test and it is not selected with automatic tests
        """
        pass



import datetime
import pytz
from django.conf import settings
from salesforce import models, models_extend

tzinfo = pytz.utc if settings.USE_TZ else None


def now_aware_or_naive():
    if settings.USE_TZ:
        return pytz.utc.localize(datetime.datetime.utcnow())
    return datetime.datetime.utcnow()


class Account(models_extend.SalesforceModel):
    name = models.CharField(max_length=255)


class Contact(models_extend.SalesforceModel):
    last_name = models.CharField(max_length=80)
    first_name = models.CharField(max_length=40, blank=True)
    account = models.ForeignKey(Account, on_delete=models.DO_NOTHING)

    class Meta:
        db_table = 'Contact'


def get_default_account() -> Account:
    return Account.objects.order_by('pk')[0]


class TryDefaults(models.SalesforceModel):
    # this model doesn't exist in Salesforce, but it should be valid
    # it is only for coverage of code by tests
    example_str = models.CharField(default=models.DefaultedOnCreate('client'), max_length=50)
    example_bool = models.BooleanField(default=models.DefaultedOnCreate(True), db_column='ExampleBool__c')
    example_bool_2 = models.BooleanField(db_column='ExampleBool2__c', default=models.DefaultedOnCreate(False))
    example_bool_3 = models.BooleanField(db_column='ExampleBool3__c', default=models.DEFAULTED_ON_CREATE)
    example_datetime = models.DateTimeField(default=models.DefaultedOnCreate(now_aware_or_naive),
                                            db_column='ExampleDatetime__c')
    example_datetime_2 = models.DateTimeField(
        default=models.DefaultedOnCreate(datetime.datetime(2021, 3, 31, 23, 59, tzinfo=tzinfo)))
    example_date = models.DateField(default=datetime.date.today)
    example_date_2 = models.DateField(default=models.DefaultedOnCreate(datetime.date(2021, 3, 31)))
    example_foreign_key = models.ForeignKey(Account, on_delete=models.DO_NOTHING,
                                            default=models.DefaultedOnCreate(get_default_account))
    example_time = models.TimeField(default=models.DefaultedOnCreate(datetime.time(23, 59)))



from django.utils.crypto import get_random_string
from salesforce.testrunner.settings import DATABASES  # noqa # necessary for setup

SECRET_KEY = get_random_string(length=32)

INSTALLED_APPS = ['tests.test_managers']
DATABASE_ROUTERS = ["salesforce.router.ModelRouter"]
SF_LAZY_CONNECT = True
USE_TZ = True
DEFAULT_AUTO_FIELD = 'django.db.models.AutoField'  # not important for Salesforce, but for Django warnings



"""
Test that MockTestCase works in all modes, including "record"
(this must connect to servers)
"""
import unittest
from typing import List, Optional
from django.db import connections
from django.test.utils import override_settings

from salesforce.backend.test_helpers import sf_alias
from tests.test_mock.mocksf import (mock, MockJsonRequest, MockTestCase,
                                    case_safe_sf_id, check_sf_api_id, extract_ids)


@override_settings(SF_MOCK_MODE='mixed')
class TestMock(MockTestCase):
    api_version = '39.0'

    def prepare_expected(self) -> None:
        self.mock_add_expected(MockJsonRequest(
            'GET mock:///services/data/v39.0/query/?q=SELECT+Name+FROM+Contact+LIMIT+1',
            resp=('{"totalSize": 1, "done": true, "records": [{'
                  '  "attributes": {"type": "Contact",'
                  '                 "url": "/services/data/v39.0/sobjects/Contact/003A000000wJICkIAO"},'
                  '  "Name": "django-salesforce test"}]}')
        ))

    @override_settings(SF_MOCK_MODE='playback')
    def test_mock_playback(self) -> None:
        self.prepare_expected()
        # test
        cur = connections[sf_alias].cursor()
        with mock.patch.object(cur.cursor.connection, '_api_version', '39.0'):
            cur.execute("SELECT Name FROM Contact LIMIT 1")
        self.assertEqual(list(cur.fetchall()), [('django-salesforce test',)])

    def test_mock_unused_playback(self) -> None:
        self.prepare_expected()
        # warning: tearDown() will be called 2x (here + autopmatic), that could cause side effects
        self.assertRaisesRegex(AssertionError, "Not all expected requests has been used", self.tearDown)

    @override_settings(SF_MOCK_MODE='record')
    @override_settings(SF_MOCK_VERBOSITY=0)
    def test_mock_record(self) -> None:
        # test
        cur = connections[sf_alias].cursor()
        cur.execute("SELECT Name FROM Contact LIMIT 1")
        row, = cur.fetchall()
        self.assertEqual(len(row), 1)
        self.assertIsInstance(row[0], str)

    @override_settings(SF_MOCK_MODE='playback')
    def test_response_parser(self) -> None:
        "Test a response parser by a Mock playback, without testing the request."
        cur = connections[sf_alias].cursor()
        req = MockJsonRequest(
            "POST mock:///services/data/v20.0/xyz",
            request_json='{"a":[123]}',
            resp="""{
                "a":3
                ...
            }""", check_request=False)
        self.mock_add_expected(req)
        ret = cur.handle_api_exceptions('POST', 'mock:///services/data/')
        self.assertEqual(ret.json()['a'], 3)


class SafeIdTest(unittest.TestCase):
    def test_case_safe_sf_id(self) -> None:
        self.assertEqual(case_safe_sf_id('000000000000000AAA'), '000000000000000AAA')
        self.assertEqual(case_safe_sf_id('000000000000000'), '000000000000000AAA')
        self.assertEqual(case_safe_sf_id('A00000000000000'), 'A00000000000000BAA')
        self.assertEqual(case_safe_sf_id(None), None)
        self.assertRaises(TypeError, case_safe_sf_id, '000000000000000AAB')
        self.assertRaises(TypeError, case_safe_sf_id, '00000000000000@')
        self.assertRaises(TypeError, case_safe_sf_id, '000000000000')

    def test_check_sf_api_id(self) -> None:
        self.assertEqual(check_sf_api_id('000000000000000'), '000000000000000AAA')
        self.assertEqual(check_sf_api_id('0'), None)

    def test_extract_ids(self) -> None:
        def extract(data: str, data_type: Optional[str] = None) -> List[str]:
            return [x[0] for x in extract_ids(data, data_type)]
        json_demo = '{"key": "000000000000000AAA", "name": "something"}'
        soap_demo = '<key>000000000000000AAA</key><name>something</name>'
        soql_demo = "WHERE key = '000000000000000AAA' AND name = 'something'"

        self.assertEqual(extract(json_demo), ['000000000000000AAA'])
        self.assertEqual(extract(soap_demo), ['000000000000000AAA'])
        self.assertEqual(extract(soql_demo), ['000000000000000AAA'])

        self.assertEqual(extract(json_demo, 'rest'), ['000000000000000AAA'])
        self.assertEqual(extract(soap_demo, 'rest'), [])
        self.assertEqual(extract(soql_demo, 'rest'), [])

        self.assertEqual(extract(json_demo, 'soap'), [])
        self.assertEqual(extract(soap_demo, 'soap'), ['000000000000000AAA'])
        self.assertEqual(extract(soql_demo, 'soap'), [])

        self.assertEqual(extract(json_demo, 'soql'), [])
        self.assertEqual(extract(soap_demo, 'soql'), [])
        self.assertEqual(extract(soql_demo, 'soql'), ['000000000000000AAA'])



from django.http import HttpResponse
from salesforce.testrunner.example import models


def account_insert_delete(request):
    # simple test view for debug-toolbar
    # Owner field is prepared by DefaultedOnCreate in the Account model
    account = models.Account(Name='abc')
    account.save()
    account.Name = 'xyz'
    account.save()
    account.delete()
    return HttpResponse('<html><body>OK</body></html>')



"""Test djangoBackward compatible behaviour with primary key 'Id'."""
import datetime
import pytz
from django.conf import settings
from django.test import TestCase

from salesforce import defaults
from .models import Account, Contact, TryDefaults


class DefaultDbTest(TestCase):
    databases = '__all__'

    def test_simple_create(self):
        """Test create new salesforce objects in a default db."""
        test_account = Account.objects.create(name='sf_test account')
        test_contact = Contact.objects.create(last_name='sf_test_contact', account=test_account)
        try:
            self.assertEqual(Account.objects.filter(name='sf_test account').count(), 1)
            self.assertEqual(Contact.objects.filter(account__name='sf_test account').count(), 1)
        finally:
            test_contact.delete()
            test_account.delete()

    def test_using_salesforce_after_default(self):
        """Test with .using('salesforce') after 'default' database queryset."""
        self.assertGreater(Contact.objects.using('salesforce').count(), 1)
        self.assertGreater(Contact.objects.all().using('salesforce').count(), 1)

    def test_using_salesforce_from_default(self):
        """Test by saving a single object save (also copy)."""
        test_account_orig = Account(name='sf_test account 0')
        test_account_orig.save(using='salesforce')
        try:
            test_account_orig.save(using='default')
            test_account_orig.save(using='salesforce')
            test_account = Account.objects.get(pk=test_account_orig.pk)
            # update salesforce
            test_account_orig.name = 'sf_test account 1'
            test_account_orig.save(using='salesforce')
            self.assertEqual(Account.objects.using('salesforce').filter(name='sf_test account 1').count(), 1)

        finally:
            test_account_orig.delete()

        # update using=default
        test_account.name = 'sf_test account 2'
        test_account.save(using='default')
        self.assertEqual(Account.objects.using('default').filter(name='sf_test account 2').count(), 1)

        # update without using
        test_account.name = 'sf_test account 3'
        test_account.save()
        self.assertEqual(Account.objects.using('default').filter(name='sf_test account 3').count(), 1)

    def test_q(self):
        test_account = Account(name='sf_test account', pk=None)
        test_account.save(using='default')
        test_account.save()  # re-save

    def test_bulk_create_alt(self):
        """Test bulk create on an alternate database"""
        # bulk_create an alternate
        objs = Account.objects.bulk_create([Account(name='sf_test account 4') for x in range(2)])
        self.assertEqual(len(objs), 2)
        test_account_qs = Account.objects.filter(name='sf_test account 4')
        self.assertEqual(test_account_qs.count(), 2)
        test_account_qs.delete()
        self.assertEqual(test_account_qs.count(), 0)

    # TODO test also bulk_create copy with  SALESFORCE_DB_ALIAS = 'salesforce'

    def test_update_delete_on_qs(self):
        """Combined tests on a default database, update and delete on a queryset"""
        test_account = Account.objects.create(name='sf_test account 0')

        # update single
        test_account.name = 'sf_test account'
        test_account.save()
        self.assertEqual(Account.objects.filter(name='sf_test account').count(), 1)

        # create with a ForeignKey and search related
        Contact.objects.create(last_name='sf_test contact', account=test_account)
        self.assertEqual(Contact.objects.filter(account__name='sf_test account').count(), 1)

        # update on queryset
        cnt = (Contact.objects.filter(account__name='sf_test account')
               .update(last_name='sf_test contact 2'))
        self.assertEqual(cnt, 1)
        self.assertEqual(Contact.objects.filter(last_name='sf_test contact 2').count(), 1)

    def test_defaulted_on_create(self):
        account = Account.objects.create(name='test')
        obj = TryDefaults.objects.create()
        obj = TryDefaults.objects.get(pk=obj.pk)
        self.assertEqual(obj.example_str, 'client')
        tzinfo = pytz.utc if settings.USE_TZ else None
        self.assertEqual(obj.example_datetime_2, datetime.datetime(2021, 3, 31, 23, 59, tzinfo=tzinfo))
        self.assertEqual(obj.example_time, datetime.time(23, 59))
        self.assertIs(obj.example_bool, True)
        self.assertIs(obj.example_bool_3, False)
        self.assertEqual(obj.example_foreign_key.pk, account.pk)
        obj2 = TryDefaults.objects.get(pk=obj.pk)  # refreshed
        self.assertEqual(obj2.example_foreign_key.pk, account.pk)
        obj.delete()
        account.delete()


class DefaultsTest(TestCase):
    def test_datetime_defaults(self) -> None:
        timezone = pytz.timezone('Europe/Prague')
        obj = defaults.DateTimeDefault(datetime.datetime(2021, 3, 31, 23, 59))
        obj2 = pytz.utc.localize(obj)
        obj2 = timezone.localize(obj)
        str(obj)
        pytz.utc.localize(obj)
        obj2.astimezone(pytz.utc)
        obj2.replace(tzinfo=None)

    def test_time_defaults(self) -> None:
        obj = defaults.TimeDefault(datetime.time(23, 59))
        self.assertTrue(isinstance(str(obj.args[0]), str))
        self.assertRegex(str(obj), r'\d{2}:\d{2}:\d{2}$')



"""Test with django-debug-toolbar."""
import datetime
from django.test import TestCase, override_settings
import django.contrib.auth
from salesforce.backend.test_helpers import uid_version as uid, expectedFailure
from salesforce.testrunner.example.models import Campaign, Contact, Lead, Opportunity


@override_settings(DEBUG=True)
class DebugToolbarViewTest(TestCase):
    """Create and delete objects in Salesforce by a simple view."""
    databases = '__all__'

    def test_account_insert_delete(self):
        resp = self.client.get('/account_insert_delete/')
        self.assertContains(resp, 'OK')


@override_settings(DEBUG=True)
class DebugToolbarAdminTest(TestCase):
    """Create objects in Salesforce by Django admin when django-debug-toolbar is enabled."""
    databases = '__all__'

    def setUp(self):
        # Log in as a superuser
        user = django.contrib.auth.models.User.objects.create_user('admin', is_superuser=True, is_staff=True)
        self.client.force_login(user)

    def test_used_debug_toolbar(self):
        """Check that django-debug-toolbar is enabled."""
        resp = self.client.get('/admin/example/campaign/add/')
        self.assertContains(resp, 'djDebugToolbar')

    def test_simple_create(self):
        resp = self.client.post('/admin/example/campaign/add/', {'name': 'test_' + uid})
        self.assertEqual(resp.status_code, 302, "Object not created")
        count_deleted, _ = Campaign.objects.filter(name='test_' + uid).delete()
        self.assertEqual(count_deleted, 1)

    def test_defaulted_numeric_create(self):
        """A simple object with a simple DefaultedOnCreate can be easily filled by web client and saved"""
        # it works becase
        resp = self.client.post(
            '/admin/example/opportunity/add/',
            dict(name='test_' + uid, stage='Prospecting', close_date=datetime.date.today())
        )
        self.assertEqual(resp.status_code, 302, "Object not created")
        opportunities = Opportunity.objects.filter(name='test_' + uid)

        # # this must fail because the DefaultedOnCreate can not pass through a web form
        # self.assertGreater(opportunities[0].probability, 0)

        count_deleted, _ = opportunities.delete()
        self.assertEqual(count_deleted, 1)

    @expectedFailure
    # This fails currently: not enough valid data for this web form because Lead object
    # has many reguired fields. This Lead object is not a good example for web tests.
    # However it works if all required fields are filled manually.
    def test_defaulted_bool_create(self):
        resp = self.client.post('/admin/example/lead/add/', dict(Company='test_' + uid, LastName='some lastname'))
        self.assertEqual(resp.status_code, 302, "Object not created")
        count_deleted, _ = Lead.objects.filter(Company='test_' + uid).delete()
        self.assertEqual(count_deleted, 1)

    @expectedFailure
    # A form with ForeignKey with DefaultedOnCreate fails, because the web widget requires
    # o concrete foreign key value and the object will be not saved.
    def test_defaulted_foreignkey_create(self):
        resp = self.client.post('/admin/example/contact/add/', dict(last_name='test_' + uid))
        self.assertEqual(resp.status_code, 302, "Object not created")
        count_deleted, _ = Contact.objects.filter(Company='test_' + uid).delete()
        self.assertEqual(count_deleted, 1)



# Generated by Django 3.2 on 2021-04-13 11:43

from django.db import migrations
import django.db.models.deletion
import django.db.models.manager
import django.utils.timezone
import salesforce.fields


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Contact',
            fields=[
                ('id', salesforce.fields.SalesforceAutoField(auto_created=True, db_column='Id', primary_key=True, serialize=False, verbose_name='ID')),
                ('last_name', salesforce.fields.CharField(max_length=40)),
                ('donor_class', salesforce.fields.CharField(blank=True, db_column='Donor_class__c', default=salesforce.fields.DefaultedOnCreate('None'), max_length=255, verbose_name='Donor class')),
            ],
            options={
                'db_table': 'Contact',
                'abstract': False,
                'base_manager_name': 'base_manager',
            },
            managers=[
                ('base_manager', django.db.models.manager.Manager()),
            ],
        ),
        migrations.CreateModel(
            name='Unreal',
            fields=[
                ('id', salesforce.fields.SalesforceAutoField(auto_created=True, db_column='Id', primary_key=True, serialize=False, verbose_name='ID')),
                ('bool_x', salesforce.fields.BooleanField(default=salesforce.fields.DefaultedOnCreate())),
                ('int_x', salesforce.fields.IntegerField(default=salesforce.fields.DefaultedOnCreate())),
                ('float_x', salesforce.fields.FloatField(default=salesforce.fields.DefaultedOnCreate())),
                ('decimal_x', salesforce.fields.DecimalField(decimal_places=2, default=salesforce.fields.DefaultedOnCreate(), max_digits=9)),
                ('str_x', salesforce.fields.CharField(default=salesforce.fields.DefaultedOnCreate(), max_length=255)),
                ('date_x', salesforce.fields.DateField(default=salesforce.fields.DefaultedOnCreate())),
                ('datetime_x', salesforce.fields.DateTimeField(default=salesforce.fields.DefaultedOnCreate())),
                ('time_x', salesforce.fields.TimeField(default=salesforce.fields.DefaultedOnCreate())),
                ('str2_x', salesforce.fields.CharField(default=salesforce.fields.DefaultedOnCreate('no'), max_length=255)),
                ('bool2_x', salesforce.fields.BooleanField(default=salesforce.fields.DefaultedOnCreate(True))),
                ('callable_timestamp', salesforce.fields.DateTimeField(default=salesforce.fields.DefaultedOnCreate(django.utils.timezone.now))),
                ('normal', salesforce.fields.CharField(max_length=255)),
            ],
            options={
                'db_table': 'Unreal',
                'abstract': False,
                'base_manager_name': 'base_manager',
            },
            managers=[
                ('base_manager', django.db.models.manager.Manager()),
            ],
        ),
        migrations.CreateModel(
            name='User',
            fields=[
                ('id', salesforce.fields.SalesforceAutoField(auto_created=True, db_column='Id', primary_key=True, serialize=False, verbose_name='ID')),
                ('name', salesforce.fields.CharField(max_length=80)),
            ],
            options={
                'db_table': 'User',
                'abstract': False,
                'base_manager_name': 'base_manager',
            },
            managers=[
                ('base_manager', django.db.models.manager.Manager()),
            ],
        ),
        migrations.CreateModel(
            name='Account',
            fields=[
                ('id', salesforce.fields.SalesforceAutoField(auto_created=True, db_column='Id', primary_key=True, serialize=False, verbose_name='ID')),
                ('name', salesforce.fields.CharField(max_length=80)),
                ('owner', salesforce.fields.ForeignKey(default=salesforce.fields.DefaultedOnCreate(), on_delete=django.db.models.deletion.DO_NOTHING, to='test_mock2.user')),
            ],
            options={
                'db_table': 'Account',
                'abstract': False,
                'base_manager_name': 'base_manager',
            },
            managers=[
                ('base_manager', django.db.models.manager.Manager()),
            ],
        ),
    ]



import unittest
import django
# in ortder to can run also by unittest without manage.py
django.setup()
from tests.inspectdb.dependent_model.models import Organization  # NOQA must be after setup()


class DependentModelTest(unittest.TestCase):
    def test_dependent_model_can_copy_fields(self):
        self.assertIn('@', Organization.objects.get().created_by.Username)
        Organization.objects.get().address


if __name__ == '__main__':
    unittest.main()



"""Backward compatible behaviour with primary key 'Id'."""
from django.test import TestCase
from .models import Contact


class AppLabelTest(TestCase):
    databases = '__all__'

    def test(self):
        obj = Contact.objects.all()[0]
        self.assertEqual(obj._meta.app_label, 'test_salesforce')



# django-salesforce
#
# by Phil Christensen
# (c) 2012-2013 Freelancers Union (http://www.freelancersunion.org)
# See LICENSE.md for details
#

import os
import re

# disables creation of .DS_Store files inside tarballs on Mac OS X
os.environ['COPY_EXTENDED_ATTRIBUTES_DISABLE'] = 'true'
os.environ['COPYFILE_DISABLE'] = 'true'


def relative_path(path):
    """
    Return the given path relative to this file.
    """
    return os.path.join(os.path.dirname(__file__), path)


def get_tagged_version():
    """
    Determine the current version of this package.

    Precise long version numbers are used if the Git repository is found.
    They contain: the Git tag, the commit serial and a short commit id.
    otherwise a short version number is used if installed from Pypi.
    """
    with open(relative_path('salesforce/__init__.py'), 'r') as fd:
        version = re.search(r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]',
                            fd.read(), re.MULTILINE).group(1)
    return version


def autosetup():
    from setuptools import setup, find_packages

    with open(relative_path('requirements.txt'), 'r', newline=None) as f:
        requirements_txt = f.read().split("\n")

    with open(relative_path('README.rst'), 'r', newline=None) as f:
        long_description = f.read()

    # check if installed with git data (not via PyPi)
    with_git = os.path.isdir(relative_path('.git'))

    return setup(
        name="django-salesforce",
        version=get_tagged_version(),

        include_package_data=True,
        zip_safe=False,
        packages=find_packages(exclude=['tests', 'tests.*']),

        # setuptools won't auto-detect Git managed files without this
        setup_requires=[] if not with_git else ["setuptools_git >= 0.4.2"],

        install_requires=['django>=2.0'] + requirements_txt,

        # metadata for upload to PyPI
        author="Freelancers Union",
        author_email="devs@freelancersunion.org",
        maintainer="Phil Christensen",
        maintainer_email="phil@bubblehouse.org",
        description="a Salesforce backend for Django's ORM",
        long_description=long_description,
        long_description_content_type="text/x-rst",
        license="MIT",
        keywords="django salesforce orm backend",
        url="https://github.com/django-salesforce/django-salesforce",
    )


if __name__ == '__main__':
    dist = autosetup()



from django.utils.crypto import get_random_string
from salesforce.testrunner.settings import DATABASES  # noqa

# DATABASES = {'default': DATABASES['salesforce']}
SECRET_KEY = get_random_string(length=32)
# SALESFORCE_DB_ALIAS = 'default'
# ROOT_URLCONF = None
INSTALLED_APPS = [
    'salesforce',
    'tests.test_mock2',
]
# DATABASE_ROUTERS = ["salesforce.router.ModelRouter"]
SF_LAZY_CONNECT = True
# USE_TZ = True
DEFAULT_AUTO_FIELD = 'django.db.models.AutoField'  # not important for Salesforce, but for Django warnings



"""Tests by parsing the file modules.py exported by inspectdb

test of a dependent module.
"""
from collections import OrderedDict
from typing import Dict
import os
import re
import unittest
from salesforce.backend import DJANGO_30_PLUS
from salesforce.backend.test_helpers import expectedFailureIf


def relative_path(path: str) -> str:
    """
    Return the given path relative to this file.
    """
    return os.path.join(os.path.dirname(__file__), path)


def get_classes_texts() -> Dict[str, str]:
    """
    Get classes texts as a dict.
    """
    result = OrderedDict()
    excluded_pattern = re.compile(r'^('
                                  r'# coding: utf-8|'
                                  r'# This is an auto-generated Django model|'
                                  r'from salesforce import models$|'
                                  r"# Unable to inspect table 'django_migrations'"
                                  r')')
    with open(relative_path('models.py'), 'r', newline=None) as f:
        for text in f.read().split('\n\n'):
            text = text.strip()
            if text and not excluded_pattern.match(text):
                match = re.match(r'class (\w+)\(', text)
                assert match, 'Unexpected line: %s' % text.split('\n')[0]
                class_name = match.groups()[0]
                result[class_name] = text
    return result


class ExportedModelTest(unittest.TestCase):

    def match_line(self, pattern, text) -> str:
        """requires the pattern and finds the line"""
        self.assertRegex(text, pattern)
        (ret,) = [line for line in text.split('\n') if re.match(pattern, line)]
        return ret

    def test_nice_fields_names(self) -> None:
        """Test the typical nice field name 'last_modified_date'."""
        for text in classes_texts.values():
            if re.search(r' last_modified_date = ', text):
                line = self.match_line(r'    last_modified_date = ', text)
                self.assertNotIn('db_column', line)
            else:
                self.assertNotIn('lastmodifieddate', text)
                self.assertNotIn('LastModifiedDate', text)

    def test_nice_standard_class_name(self):
        self.assertTrue('AccountContactRole' in classes_texts.keys())

    def test_custom_test_class(self) -> None:
        """Test the typical nice class name 'Test'."""
        for name, text in classes_texts.items():
            if re.search(r"        db_table = 'django_Test__c'", text):
                # test the class name
                self.assertEqual(name, 'DjangoTest')
                # test the field name without db_column
                self.assertNotIn('db_column', self.match_line(r'    test_text = ', text))
                # test foreign kea
                line = self.match_line(r'    contact = ', text)
                self.assertIn('custom=True', line)
                self.assertIn('ForeignKey(Contact', line)
                self.assertIn(', models.DO_NOTHING,', line)
        else:
            self.skipTest("The model for the table Test__c not exported.")

    def test_master_detail_relationship(self) -> None:
        """
        Verify that Contact is a master-detail relationship of Account,
        but Opportunity is not.
        """
        line = self.match_line('    account = ', classes_texts['Contact'])
        self.assertRegex(line, r'#.* Master Detail Relationship \*')
        line = self.match_line('    created_by = ', classes_texts['Opportunity'])
        self.assertNotIn('Master Detail Relationship', line)

    @expectedFailureIf(not DJANGO_30_PLUS)
    def test_one_to_one_field_introspection(self) -> None:
        """Test that OneToOneField is correctly introspected"""
        self.assertTrue(any(('= models.OneToOneField(' in text) for text in classes_texts.values()))


classes_texts = get_classes_texts()

if __name__ == '__main__':
    unittest.main()



from salesforce.testrunner.settings import *  # NOQA pylint: disable=unused-wildcard-import,wildcard-import
from salesforce.testrunner.settings import DATABASES, INSTALLED_APPS

SF_LAZY_CONNECT = True
DATABASES = {'default': DATABASES['default'],
             'salesforce': DATABASES['salesforce']}
ORIG_SALESFORCE_DB = DATABASES['salesforce'].copy()
DATABASES['salesforce'].update(HOST='https://nonsense.example.com',
                               CONSUMER_KEY='.',
                               CONSUMER_SECRET='.',
                               USER='.',
                               PASSWORD='.')
INSTALLED_APPS += ('tests.test_lazy_connect',)
